# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload

import operator
from collections.abc import Iterator
from datetime import datetime
from datetime import timedelta
from datetime import UTC

from gi.repository import Gdk
from gi.repository import GLib
from nbxmpp import unescape_localpart
from nbxmpp.const import Affiliation
from nbxmpp.const import Chatstate
from nbxmpp.const import PresenceShow
from nbxmpp.const import Role
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import HatData
from nbxmpp.structs import LocationData
from nbxmpp.structs import MucSubject
from nbxmpp.structs import TuneData

import gajim.common.storage.archive.models as mod
from gajim.common import app
from gajim.common import modules
from gajim.common import types
from gajim.common.const import PresenceShowExt
from gajim.common.const import SimpleClientState
from gajim.common.const import VALUE_MISSING
from gajim.common.const import ValueMissingT
from gajim.common.helpers import Observable
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import LogAdapter
from gajim.common.setting_values import AllContactSettings
from gajim.common.setting_values import AllContactSettingsT
from gajim.common.setting_values import AllGroupChatSettings
from gajim.common.setting_values import AllGroupChatSettingsT
from gajim.common.setting_values import BoolContactSettings
from gajim.common.setting_values import BoolGroupChatSettings
from gajim.common.setting_values import IntGroupChatSettings
from gajim.common.setting_values import StringContactSettings
from gajim.common.setting_values import StringGroupChatSettings
from gajim.common.structs import MUCPresenceData
from gajim.common.structs import PresenceData
from gajim.common.structs import UNKNOWN_MUC_PRESENCE
from gajim.common.structs import UNKNOWN_PRESENCE
from gajim.common.util.muc import get_groupchat_name
from gajim.common.util.user_strings import chatstate_to_string


class ContactSettings:
    def __init__(self, account: str, jid: JID) -> None:
        self._account = account
        self._jid = jid

    @overload
    def get(self, setting: StringContactSettings) -> str: ...
    @overload
    def get(self, setting: BoolContactSettings) -> bool: ...

    def get(self, setting: AllContactSettings) -> AllContactSettingsT:
        return app.settings.get_contact_setting(self._account, self._jid, setting)

    @overload
    def set(self, setting: StringContactSettings, value: str | None) -> None: ...  # noqa: E501
    @overload
    def set(self, setting: BoolContactSettings, value: bool | None) -> None: ...  # noqa: E501

    def set(self, setting: Any, value: Any) -> None:
        app.settings.set_contact_setting(self._account, self._jid, setting, value)


class GroupChatSettings:
    def __init__(self, account: str, jid: JID) -> None:
        self._account = account
        self._jid = jid

    @overload
    def get(self, setting: StringGroupChatSettings) -> str: ...
    @overload
    def get(self, setting: BoolGroupChatSettings) -> bool: ...
    @overload
    def get(self, setting: IntGroupChatSettings) -> int: ...

    def get(self, setting: AllGroupChatSettings) -> AllGroupChatSettingsT:
        return app.settings.get_group_chat_setting(self._account, self._jid, setting)

    @overload
    def set(self, setting: StringGroupChatSettings, value: str | None) -> None: ...  # noqa: E501
    @overload
    def set(self, setting: BoolGroupChatSettings, value: bool | None) -> None: ...  # noqa: E501
    @overload
    def set(self, setting: IntGroupChatSettings, value: int | None) -> None: ...  # noqa: E501

    def set(self, setting: Any, value: Any) -> None:
        app.settings.set_group_chat_setting(self._account, self._jid, setting, value)


class Contacts(BaseModule):
    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

        self._contacts: dict[JID, BareContact | GroupchatContact] = {}
        self._con.connect_signal("state-changed", self._on_client_state_changed)
        self._con.connect_signal("resume-failed", self._on_client_resume_failed)

    def _on_client_resume_failed(
        self, _client: types.Client, _signal_name: str
    ) -> None:
        self._reset_presence()

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        if state.is_disconnected:
            self._reset_presence()

    def add_chat_contact(self, jid: str | JID) -> BareContact:
        if isinstance(jid, str):
            jid = JID.from_string(jid)

        contact = self._contacts.get(jid)
        if contact is not None:
            if not isinstance(contact, BareContact):
                raise ValueError(
                    f"Trying to add BareContact {jid}, "
                    f"but contact exists already as {contact}"
                )
            return contact

        contact = BareContact(self._log, jid, self._account)

        self._contacts[jid] = contact
        return contact

    def add_group_chat_contact(self, jid: str | JID) -> GroupchatContact:
        if isinstance(jid, str):
            jid = JID.from_string(jid)

        contact = self._contacts.get(jid)
        if contact is not None:
            if isinstance(contact, GroupchatContact):
                return contact

            self._log.error(
                f"Trying to add GroupchatContact {jid}, "
                f"but contact already exists as {contact} "
                f"(in roster: {contact.is_in_roster})"
            )
            # Most likely a groupchat which has been added to the roster.
            # Remove jid from roster (if we are online) and delete contact.
            if self._client.is_available():
                self._client.get_module("Roster").delete_item(jid)
            self._contacts.pop(jid)

        contact = GroupchatContact(self._log, jid, self._account)

        self._contacts[jid] = contact
        return contact

    def add_private_contact(self, jid: str | JID) -> GroupchatParticipant:
        if isinstance(jid, str):
            jid = JID.from_string(jid)

        if jid.resource is None:
            raise ValueError(f"Trying to add a bare JID as private {jid}")

        contact = self._contacts.get(jid.new_as_bare())
        if contact is None:
            group_chat_contact = self.add_group_chat_contact(jid.bare)
            return group_chat_contact.add_resource(jid.resource)

        if not isinstance(contact, GroupchatContact):
            raise ValueError(
                f"Trying to add GroupchatParticipant {jid}, to BareContact {contact}"
            )

        return contact.add_resource(jid.resource)

    def add_contact(
        self, jid: str | JID, groupchat: bool = False
    ) -> BareContact | GroupchatContact:
        if isinstance(jid, str):
            jid = JID.from_string(jid)

        contact = self._contacts.get(jid)
        if contact is not None:
            return contact

        if groupchat:
            contact = GroupchatContact(self._log, jid, self._account)
        else:
            contact = BareContact(self._log, jid, self._account)

        self._contacts[jid] = contact
        return contact

    def get_contact(self, jid: str | JID, groupchat: bool = False) -> types.ContactT:

        if isinstance(jid, str):
            jid = JID.from_string(jid)

        resource = jid.resource
        jid = jid.new_as_bare()

        contact = self._contacts.get(jid)
        if contact is None:
            contact = self.add_contact(jid, groupchat=groupchat)

        if resource is None:
            return contact

        contact = contact.get_resource(resource)
        return contact

    def get_contact_if_exists(
        self, jid: JID
    ) -> BareContact | ResourceContact | GroupchatContact | GroupchatParticipant | None:
        contact = self._contacts.get(jid.new_as_bare())
        if contact is None:
            return None

        if jid.resource:
            return contact.get_resource_if_exists(jid.resource)
        return contact

    def get_contacts_with_domain(
        self, domain: str
    ) -> list[BareContact | GroupchatContact]:

        contacts: list[BareContact | GroupchatContact] = []
        for contact in self._contacts.values():
            if contact.jid.domain == domain:
                contacts.append(contact)
        return contacts

    def _reset_presence(self) -> None:
        for contact in self._contacts.values():
            if contact.is_groupchat or contact.is_pm_contact:
                continue
            assert isinstance(contact, BareContact)
            contact.update_presence(UNKNOWN_PRESENCE)

    def force_chatstate_update(self) -> None:
        for contact in self._contacts.values():
            contact.force_chatstate_update()


class CommonContact(Observable):
    def __init__(self, logger: LogAdapter | None, jid: JID, account: str) -> None:

        Observable.__init__(self, logger)
        self._jid = jid
        self._account = account
        self._gateway_type: str | None = None

    def __hash__(self) -> int:
        return hash(f"{self._account}-{self._jid}")

    def __eq__(self, obj: object) -> bool:
        if not isinstance(obj, CommonContact):
            return NotImplemented

        return self._account == obj.account and obj._jid == self._jid

    @property
    def jid(self) -> JID:
        return self._jid

    @property
    def account(self) -> str:
        return self._account

    def get_module(self, name: modules.ModulesLiteralT) -> modules.ModulesT:
        return modules.get_module(self._account, name)

    def _on_signal(
        self,
        _contact: (
            BareContact | ResourceContact | GroupchatContact | GroupchatParticipant
        ),
        signal_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.notify(signal_name, *args, **kwargs)

    def supports(self, requested_feature: str) -> bool:
        return self._supports(self._jid, requested_feature)

    @staticmethod
    def _supports(jid: JID, requested_feature: str) -> bool:
        disco_info = app.storage.cache.get_last_disco_info(jid)
        if disco_info is None:
            return False

        return disco_info.supports(requested_feature)

    @property
    def is_chat(self) -> bool:
        return False

    @property
    def is_groupchat(self) -> bool:
        return False

    @property
    def is_pm_contact(self) -> bool:
        return False

    @property
    def type_string(self) -> str:
        raise NotImplementedError

    @property
    def is_jingle_available(self) -> bool:
        return False

    def get_address(self, _prefer_real: bool = True) -> JID:
        return self._jid

    @property
    def chatstate(self) -> Chatstate | None:
        return None

    @property
    def chatstate_string(self) -> str:
        return chatstate_to_string(self.chatstate)

    @property
    def is_muted(self) -> bool:
        mute_until = self.settings.get("mute_until")
        if not mute_until:
            return False

        until = datetime.fromisoformat(mute_until)
        is_muted = until > datetime.now(UTC)
        if not is_muted:
            # Reset the setting to default
            GLib.idle_add(self.settings.set, "mute_until", None)
        return is_muted

    def __repr__(self) -> str:
        return f"{self.jid} ({self._account})"

    def _get_transport_icon_name(self) -> str | None:
        if self._gateway_type is not None:
            return f"gateway-{self._gateway_type}"

        domain_disco = app.storage.cache.get_last_disco_info(self._jid.domain)
        if domain_disco is None:
            return None

        if gateway_type := domain_disco.gateway_type:
            self._gateway_type = gateway_type
            return f"gateway-{gateway_type}"

    def update_gateway_type(self, gateway_type: str | None):
        if gateway_type == self._gateway_type:
            return

        self._gateway_type = gateway_type
        self.notify("avatar-update")


class BareContact(CommonContact):
    def __init__(self, logger: LogAdapter, jid: JID, account: str) -> None:
        CommonContact.__init__(self, logger, jid, account)

        self.settings = ContactSettings(account, jid)
        self._resources: dict[str, ResourceContact] = {}

        self._avatar_sha = app.storage.archive.get_contact_value(
            account, jid, "avatar_sha"
        )
        self._presence = UNKNOWN_PRESENCE

    @property
    def is_self(self):
        own_jid = app.get_client(self._account).get_own_jid().new_as_bare()
        return own_jid == self.jid

    def supports(self, requested_feature: str) -> bool:
        if not self._resources:
            return super().supports(requested_feature)
        return any(
            resource.supports(requested_feature) for resource in self.iter_resources()
        )

    @property
    def supports_audio(self) -> bool:
        if self.supports(Namespace.JINGLE_ICE_UDP) and app.is_installed("FARSTREAM"):
            return self.supports(Namespace.JINGLE_RTP_AUDIO)
        return False

    @property
    def supports_video(self) -> bool:
        if self.supports(Namespace.JINGLE_ICE_UDP) and app.is_installed("FARSTREAM"):
            return self.supports(Namespace.JINGLE_RTP_VIDEO)
        return False

    def add_resource(self, resource: str) -> ResourceContact:
        assert resource is not None
        # Check if resource is not None because not the whole
        # codebase is type checked and it creates hard to track
        # problems if we create a ResourceContact without resource

        jid = self._jid.new_with(resource=resource)
        assert isinstance(self._log, LogAdapter)
        contact = ResourceContact(self._log, jid, self._account)
        self._resources[resource] = contact
        contact.connect("presence-update", self._on_signal)
        contact.connect("nickname-update", self._on_signal)
        contact.connect("caps-update", self._on_signal)
        return contact

    def has_resources(self) -> bool:
        return bool(self._resources)

    def get_resource(self, resource: str) -> ResourceContact:
        contact = self._resources.get(resource)
        if contact is None:
            contact = self.add_resource(resource)
        return contact

    def get_resource_if_exists(self, resource: str) -> ResourceContact | None:
        return self._resources.get(resource)

    def get_active_resource(self) -> ResourceContact | None:
        if not self._resources:
            return None

        return max(self._resources.values(), key=operator.attrgetter("show"))

    def get_resources(self) -> list[ResourceContact]:
        resources: list[ResourceContact] = []
        for contact in self._resources.values():
            if contact.is_available:
                resources.append(contact)
        return resources

    def iter_resources(self) -> Iterator[ResourceContact]:
        for contact in self._resources.values():
            if contact.is_available:
                yield contact

    @property
    def is_available(self) -> bool:
        if not self._resources:
            return self._presence.available
        return any(contact.is_available for contact in self._resources.values())

    @property
    def show(self) -> PresenceShow | Literal[PresenceShowExt.OFFLINE]:
        if not self._resources:
            if not self._presence.available:
                return PresenceShowExt.OFFLINE
            return self._presence.show

        res = self.get_active_resource()
        assert res is not None
        return res.show

    @property
    def status(self) -> str:
        if not self._resources:
            return self._presence.status

        res = self.get_active_resource()
        assert res is not None
        return res.status

    @property
    def idle_datetime(self) -> datetime | None:
        if not self._resources:
            return self._presence.idle_datetime

        res = self.get_active_resource()
        assert res is not None
        return res.idle_datetime

    @property
    def chatstate(self) -> Chatstate | None:
        return self.get_module("Chatstate").get_remote_chatstate(self._jid)

    def force_chatstate_update(self) -> None:
        self.notify("chatstate-update")

    @property
    def name(self) -> str:
        if custom_name := app.storage.archive.get_contact_value(
            self._account, self._jid, "custom_name"
        ):
            return custom_name

        return self.original_name

    @property
    def original_name(self) -> str:
        if remote_name := app.storage.archive.get_contact_value(
            self._account, self._jid, "remote_name"
        ):
            return remote_name

        if self._jid.is_domain:
            assert self._jid.domain is not None
            return self._jid.domain

        assert self._jid.localpart is not None
        return unescape_localpart(self._jid.localpart)

    def get_tune(self) -> TuneData | None:
        return self.get_module("UserTune").get_contact_tune(self._jid)

    def get_location(self) -> LocationData | None:
        return self.get_module("UserLocation").get_contact_location(self._jid)

    @property
    def avatar_sha(self) -> str | None:
        return self._avatar_sha

    def get_avatar(
        self,
        size: int,
        scale: int,
        add_show: bool = True,
        default: bool = False,
        style: str = "circle",
    ) -> Gdk.Texture:

        show = self.show.value if add_show else None

        transport_icon = self._get_transport_icon_name()

        return app.app.avatar_storage.get_texture(
            self,
            size,
            scale,
            show,
            default=default,
            transport_icon=transport_icon,
            style=style,
        )

    def update_presence(self, presence_data: PresenceData) -> None:
        self._presence = presence_data
        if not presence_data.available:
            for contact in self._resources.values():
                contact.update_presence(presence_data, notify=False)
        self.notify("presence-update")

    def update_avatar(self, sha: str | None) -> None:
        if self._avatar_sha == sha:
            return

        self._avatar_sha = sha

        app.app.avatar_storage.invalidate_cache(self._jid)
        self.notify("avatar-update")

    @property
    def is_in_roster(self) -> bool:
        item = self.get_module("Roster").get_item(self._jid)
        return item is not None

    @property
    def is_gateway(self) -> bool:
        disco_info = app.storage.cache.get_last_disco_info(self._jid)
        if disco_info is None:
            return False
        return disco_info.is_gateway

    @property
    def ask(self) -> str | None:
        item = self.get_module("Roster").get_item(self._jid)
        if item is None:
            return None
        return item.ask

    @property
    def subscription(self) -> str | None:
        item = self.get_module("Roster").get_item(self._jid)
        if item is None:
            return None
        return item.subscription

    @property
    def groups(self) -> set[str]:
        item = self.get_module("Roster").get_item(self._jid)
        if item is None:
            return set()
        return item.groups

    @property
    def is_subscribed(self) -> bool:
        return self.subscription in ("from", "both")

    @property
    def is_blocked(self) -> bool:
        return self.get_module("Blocking").is_blocked(self._jid)

    def set_blocked(self) -> None:
        self.update_presence(UNKNOWN_PRESENCE)
        self.notify("blocking-update")

    def set_unblocked(self) -> None:
        self.notify("blocking-update")

    @property
    def is_jingle_available(self) -> bool:
        if not self.supports(Namespace.JINGLE_FILE_TRANSFER_5):
            return False
        return self.is_available

    @property
    def is_chat(self) -> bool:
        return True

    @property
    def type_string(self) -> Literal["chat"]:
        return "chat"

    @property
    def is_bot(self) -> bool:
        disco_info = app.storage.cache.get_last_disco_info(self._jid)

        if disco_info is not None and disco_info.has_identity("client", "bot"):
            return True

        return any(r.is_bot for r in self.iter_resources())

    @property
    def reactions_per_user(self) -> int | None:
        disco = app.storage.cache.get_last_disco_info(self._jid.domain)
        if disco is None:
            return None
        return disco.reactions_per_user


class ResourceContact(CommonContact):
    def __init__(self, logger: LogAdapter, jid: JID, account: str) -> None:
        CommonContact.__init__(self, logger, jid, account)

        self._presence = UNKNOWN_PRESENCE

    def supports(self, requested_feature: str) -> bool:
        if not self.is_available:
            return False
        return CommonContact.supports(self, requested_feature)

    @property
    def resource(self) -> str:
        assert self._jid.resource is not None
        return self._jid.resource

    @property
    def identity_type(self) -> str | None:
        disco_info = app.storage.cache.get_last_disco_info(self._jid)
        if disco_info is None:
            return None

        for identity in disco_info.identities:
            if identity.type is not None:
                return identity.type

        return None

    @property
    def is_phone(self):
        disco_info = app.storage.cache.get_last_disco_info(self._jid)
        if disco_info is None:
            return False

        return any(identity.type == "phone" for identity in disco_info.identities)

    @property
    def is_bot(self):
        disco_info = app.storage.cache.get_last_disco_info(self._jid)
        return disco_info is not None and disco_info.has_identity("client", "bot")

    @property
    def is_available(self) -> bool:
        return self._presence.available

    @property
    def show(self):
        if not self._presence.available:
            return PresenceShowExt.OFFLINE
        return self._presence.show

    @property
    def status(self) -> str:
        return self._presence.status

    @property
    def priority(self) -> int:
        return self._presence.priority

    @property
    def idle_datetime(self) -> datetime | None:
        return self._presence.idle_datetime

    def update_presence(self, presence_data: PresenceData, notify: bool = True) -> None:
        self._presence = presence_data
        if notify:
            self.notify("presence-update")

    @property
    def type_string(self) -> str:
        raise NotImplementedError

    @property
    def is_muted(self) -> bool:
        raise NotImplementedError


class GroupchatContact(CommonContact):
    def __init__(self, logger: LogAdapter, jid: JID, account: str) -> None:
        CommonContact.__init__(self, logger, jid, account)

        self.settings = GroupChatSettings(account, jid)
        self._resources: dict[str, GroupchatParticipant] = {}

        self._avatar_sha = app.storage.archive.get_contact_value(
            account, jid, "avatar_sha"
        )

    @property
    def is_groupchat(self) -> bool:
        return True

    @property
    def is_irc(self) -> bool:
        disco_info = self.get_disco()
        if disco_info is None:
            return False
        return disco_info.is_irc

    @property
    def muc_context(self) -> str | None:
        disco_info = self.get_disco()
        if disco_info is None:
            return None

        if disco_info.muc_is_members_only and disco_info.muc_is_nonanonymous:
            return "private"
        return "public"

    @property
    def encryption_available(self) -> bool:
        disco_info = self.get_disco()
        if disco_info is None:
            return True

        return disco_info.muc_is_members_only and disco_info.muc_is_nonanonymous

    def get_config_value(self, field_name: str) -> Any:
        disco_info = self.get_disco()
        assert disco_info is not None
        return disco_info.get_field_value(Namespace.MUC_INFO, field_name)

    def add_resource(self, resource: str) -> GroupchatParticipant:
        assert resource is not None
        # Check if resource is not None because not the whole
        # codebase is type checked and it creates hard to track
        # problems if we create a GroupchatParticipant without resource

        contact = self._resources.get(resource)
        if contact is not None:
            return contact

        jid = self._jid.new_with(resource=resource)
        assert isinstance(self._log, LogAdapter)
        contact = GroupchatParticipant(self._log, jid, self._account)
        self._resources[resource] = contact
        contact.connect("user-joined", self._on_user_signal)
        contact.connect("user-left", self._on_user_signal)
        contact.connect("user-affiliation-changed", self._on_user_signal)
        contact.connect("user-role-changed", self._on_user_signal)
        contact.connect("user-hats-changed", self._on_user_signal)
        contact.connect("user-status-show-changed", self._on_user_signal)
        contact.connect("user-avatar-update", self._on_user_signal)
        return contact

    def get_resource(self, resource: str) -> GroupchatParticipant:
        contact = self._resources.get(resource)
        if contact is None:
            contact = self.add_resource(resource)
        return contact

    def get_resource_if_exists(self, resource: str) -> GroupchatParticipant | None:
        return self._resources.get(resource)

    def get_participants(self) -> Iterator[GroupchatParticipant]:
        for contact in self._resources.values():
            if contact.is_available:
                yield contact

    def get_occupant(self, occupant_id: str) -> GroupchatParticipant | None:
        for contact in self._resources.values():
            if contact.occupant_id == occupant_id:
                return contact
        return None

    @property
    def name(self) -> str:
        client = app.get_client(self._account)
        return get_groupchat_name(client, self._jid)

    @property
    def original_name(self) -> str:
        assert self._jid.localpart is not None
        original_name = self._jid.localpart

        disco_info = app.storage.cache.get_last_disco_info(self._jid)
        if disco_info is not None:
            if disco_info.muc_name:
                original_name = disco_info.muc_name

        return original_name

    @property
    def avatar_sha(self) -> str | None:
        return self._avatar_sha

    def get_avatar(self, size: int, scale: int) -> Gdk.Texture:
        transport_icon = self._get_transport_icon_name()

        return app.app.avatar_storage.get_muc_texture(
            self._account, self._jid, size, scale, transport_icon=transport_icon
        )

    def _on_user_signal(
        self, contact: GroupchatParticipant, signal_name: str, *args: Any
    ) -> None:
        self.notify(signal_name, contact, *args)

    def update_avatar(self, sha: str | None) -> None:
        if self._avatar_sha == sha:
            return

        self._avatar_sha = sha

        app.app.avatar_storage.invalidate_cache(self._jid)
        self.notify("avatar-update")

    def force_chatstate_update(self) -> None:
        for contact in self._resources.values():
            contact.notify("chatstate-update")

    def get_self(self) -> GroupchatParticipant | None:
        nick = self.nickname
        if nick is None:
            return None
        return self.get_resource(nick)

    def get_offline_members(
        self, affiliation: str | None = None
    ) -> Iterator[GroupchatOfflineParticipant]:

        participant_real_jids = {
            c.real_jid for c in self.get_participants() if c.real_jid is not None
        }

        affiliations = self.get_module("MUC").get_affiliations(self._jid)
        for name, jids in affiliations.items():
            if not jids:
                continue

            if affiliation is not None and name != affiliation:
                continue

            offline_jids = set(jids) - participant_real_jids
            occupants = app.storage.archive.get_occupant_by_jids(
                self._account, self._jid, offline_jids, max_age=timedelta(minutes=60)
            )

            for jid in offline_jids:
                yield GroupchatOfflineParticipant(
                    self._account, jid, self, name, occupants.get(jid)
                )

    @property
    def nickname(self) -> str | None:
        muc_data = self.get_module("MUC").get_muc_data(self._jid)
        if muc_data is None:
            return None
        return muc_data.nick

    @property
    def occupant_jid(self) -> JID | None:
        muc_data = self.get_module("MUC").get_muc_data(self._jid)
        if muc_data is None:
            return None
        return muc_data.occupant_jid

    @property
    def subject(self) -> MucSubject | None:
        muc_data = self.get_module("MUC").get_muc_data(self._jid)
        if muc_data is None:
            return None
        return muc_data.subject

    @property
    def is_joined(self) -> bool:
        muc_data = self.get_module("MUC").get_muc_data(self._jid)
        if muc_data is None:
            return False
        return muc_data.state.is_joined

    @property
    def is_joining(self) -> bool:
        muc_data = self.get_module("MUC").get_muc_data(self._jid)
        if muc_data is None:
            return False
        return muc_data.state.is_joining or muc_data.state.is_before_joining

    @property
    def is_not_joined(self) -> bool:
        muc_data = self.get_module("MUC").get_muc_data(self._jid)
        if muc_data is None:
            return True
        return muc_data.state.is_not_joined

    def set_not_joined(self) -> None:
        for contact in self._resources.values():
            contact.update_presence(UNKNOWN_MUC_PRESENCE)

    def get_user_nicknames(self) -> list[str]:
        client = app.get_client(self._account)
        return client.get_module("MUC").get_joined_users(self._jid)

    def get_disco(self, max_age: int = 0) -> DiscoInfo | None:
        return app.storage.cache.get_last_disco_info(self.jid, max_age=max_age)

    def can_notify(self) -> bool:
        all_ = app.settings.get("notify_on_all_muc_messages")
        room = self.settings.get("notify_on_all_messages")
        return all_ or room

    @property
    def type_string(self) -> Literal["groupchat"]:
        return "groupchat"

    def has_composing_participants(self) -> bool:
        return bool(self.get_module("Chatstate").get_composers(self._jid))

    def get_composers(self) -> list["GroupchatParticipant"]:
        return self.get_module("Chatstate").get_composers(self._jid)

    @property
    def reactions_per_user(self) -> int | None:
        for entity in (self._jid, self._jid.domain):
            disco = app.storage.cache.get_last_disco_info(entity)
            if disco is None:
                continue

            if disco.reactions_per_user is not None:
                return disco.reactions_per_user

        return None


class GroupchatParticipant(CommonContact):
    def __init__(self, logger: LogAdapter, jid: JID, account: str) -> None:
        CommonContact.__init__(self, logger, jid, account)

        self.settings = ContactSettings(account, jid)

        self._client = app.get_client(self._account)

        self._presence = UNKNOWN_MUC_PRESENCE

    @property
    def resource(self) -> str:
        assert self._jid.resource is not None
        return self._jid.resource

    def get_address(self, prefer_real: bool = True) -> JID:
        jid = self._presence.real_jid
        if jid is None or not prefer_real:
            return self._jid
        return jid

    def supports(self, requested_feature: str) -> bool:
        if not self.is_available:
            return False
        return CommonContact.supports(self, requested_feature)

    @property
    def is_pm_contact(self) -> bool:
        return True

    @property
    def is_in_roster(self) -> bool:
        return False

    @property
    def room(self) -> GroupchatContact:
        contact = self.get_module("Contacts").get_contact_if_exists(
            self.jid.new_as_bare()
        )
        assert isinstance(contact, GroupchatContact)
        return contact

    @property
    def muc_context(self) -> str | None:
        return self.room.muc_context

    @property
    def presence(self) -> MUCPresenceData:
        return self._presence

    def set_presence(self, presence: MUCPresenceData) -> None:
        self._presence = presence

    @property
    def is_available(self) -> bool:
        return self._presence.available

    @property
    def show(self):
        if not self._presence.available:
            return PresenceShowExt.OFFLINE
        return self._presence.show

    @property
    def status(self) -> str:
        return self._presence.status

    @property
    def idle_datetime(self) -> datetime | None:
        return self._presence.idle_datetime

    @property
    def name(self) -> str:
        assert self._jid.resource is not None
        return self._jid.resource

    @property
    def real_jid(self) -> JID | None:
        return self._presence.real_jid

    def get_real_contact(self) -> BareContact | None:
        jid = self._presence.real_jid
        if jid is None:
            return None
        contact = self._client.get_module("Contacts").get_contact(jid.new_as_bare())
        assert isinstance(contact, BareContact)
        return contact

    @property
    def affiliation(self) -> Affiliation:
        return self._presence.affiliation

    @property
    def role(self) -> Role:
        return self._presence.role

    @property
    def chatstate(self) -> Chatstate | None:
        return self.get_module("Chatstate").get_remote_chatstate(self._jid)

    @property
    def avatar_sha(self) -> str | None:
        return self._client.get_module("VCardAvatars").get_avatar_sha(self._jid)

    def get_avatar(
        self, size: int, scale: int, add_show: bool = True, style: str = "circle"
    ) -> Gdk.Texture:

        show = self.show.value if add_show else None
        return app.app.avatar_storage.get_texture(self, size, scale, show, style=style)

    def update_presence(self, presence: MUCPresenceData) -> None:
        self._presence = presence

    def update_avatar(self, *args: Any) -> None:
        app.app.avatar_storage.invalidate_cache(self._jid)
        self.notify("user-avatar-update")

    @property
    def type_string(self) -> Literal["pm"]:
        return "pm"

    @property
    def occupant_id(self) -> str | None:
        return self._presence.occupant_id

    @property
    def is_self(self) -> bool:
        data = self.get_module("MUC").get_muc_data(self.room.jid)
        assert data is not None
        return data.nick == self.name

    @property
    def hats(self) -> HatData | None:
        return self._presence.hats

    @property
    def is_blocked(self) -> bool:
        client = app.get_client(self._account)
        return client.get_module("MucBlocking").is_blocked(
            self.room.jid, self.occupant_id
        )

    @property
    def reactions_per_user(self) -> int | None:
        return self.room.reactions_per_user


class GroupchatOfflineParticipant(CommonContact):
    def __init__(
        self,
        account: str,
        jid: JID,
        room: GroupchatContact,
        affiliation: str,
        occupant: mod.Occupant | ValueMissingT | None = VALUE_MISSING,
    ) -> None:
        CommonContact.__init__(self, None, jid, account)

        self._affiliation = Affiliation[affiliation.upper()]
        self._room = room
        self._client = app.get_client(self._account)

        if isinstance(occupant, ValueMissingT):
            occupants = app.storage.archive.get_occupant_by_jids(
                account, room.jid, [jid], max_age=timedelta(minutes=60)
            )
            self._occupant = occupants.get(jid)
        else:
            self._occupant = occupant

    @classmethod
    def from_participant(
        cls, contact: GroupchatParticipant
    ) -> GroupchatOfflineParticipant:
        assert contact.real_jid is not None

        return cls(
            contact.account,
            contact.real_jid,
            contact.room,
            contact.affiliation.value,
        )

    @property
    def room(self) -> GroupchatContact:
        return self._room

    @property
    def affiliation(self) -> Affiliation:
        return self._affiliation

    @property
    def role(self) -> Role:
        return Role.NONE

    @property
    def is_available(self) -> bool:
        return False

    @property
    def name(self) -> str:
        if self._occupant is not None:
            if nickname := self._occupant.nickname:
                return nickname
        return str(self.jid)

    @property
    def real_jid(self) -> JID:
        return self.jid

    @property
    def hats(self) -> None:
        return None

    @property
    def status(self) -> str:
        return ""

    @property
    def show(self) -> PresenceShowExt:
        return PresenceShowExt.OFFLINE

    @property
    def occupant_id(self) -> str | None:
        if self._occupant is not None:
            return self._occupant.id
        return None

    @property
    def is_self(self) -> bool:
        return False

    @property
    def avatar_sha(self) -> str | None:
        if self._occupant is not None:
            if sha := self._occupant.avatar_sha:
                return sha
        return None

    def get_avatar(
        self, size: int, scale: int, add_show: bool = True, style: str = "circle"
    ) -> Gdk.Texture:

        return app.app.avatar_storage.get_texture(
            self, size, scale, self.show.value, style=style
        )

    def get_real_contact(self) -> BareContact:
        contact = self._client.get_module("Contacts").get_contact(self._jid)
        assert isinstance(contact, BareContact)
        return contact

    @property
    def is_blocked(self) -> bool:
        client = app.get_client(self._account)
        return client.get_module("MucBlocking").is_blocked(
            self._room.jid, self.occupant_id
        )


def can_add_to_roster(contact: BareContact) -> bool:
    if contact.is_self:
        return False
    return not contact.is_in_roster



from __future__ import annotations

import logging
import uuid

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common import types
from gajim.common.const import CallType
from gajim.common.const import JingleState
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.jingle_rtp import JingleAudio
from gajim.common.jingle_session import JingleSession
from gajim.common.modules.contacts import BareContact
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.datetime import utc_now

from gajim.gtk import sound

log = logging.getLogger("gajim.c.call_manager")


class CallManager(EventHelper):
    def __init__(self) -> None:
        EventHelper.__init__(self)
        self._account: str | None = None
        self._resource_jid: JID | None = None

        self._jingle_audio_sid: str | None = None
        self._jingle_video_sid: str | None = None

        self._jingle_audio_state: JingleState = JingleState.NULL
        self._jingle_video_state: JingleState = JingleState.NULL

        # Helper for upgrading voice call to voice + video without
        # having to show a new call row
        self._incoming_video_event: events.JingleRequestReceived | None = None

        # pylint: disable=line-too-long
        self.register_events(
            [
                ("jingle-request-received", ged.GUI2, self._on_jingle_request),
                ("jingle-connected-received", ged.GUI2, self._on_jingle_connected),
                (
                    "jingle-disconnected-received",
                    ged.GUI2,
                    self._on_jingle_disconnected,
                ),
                ("jingle-error-received", ged.GUI2, self._on_jingle_error),
            ]
        )
        # pylint: enable=line-too-long

    def _on_jingle_request(self, event: events.JingleRequestReceived) -> None:
        content_types: list[str] = []
        for item in event.contents:
            content_types.append(item.media)
        if not any(item in ("audio", "video") for item in content_types):
            return

        if self._resource_jid is not None:
            if str(self._resource_jid) != event.fjid:
                log.info(
                    "Received jingle request while in active call: %s, %s",
                    event.account,
                    event.jid,
                )
                return

        self._account = event.account
        self._resource_jid = JID.from_string(event.fjid)

        if "audio" in content_types:
            self._set_jingle_state("audio", JingleState.CONNECTION_RECEIVED, event.sid)
        if "video" in content_types:
            self._set_jingle_state("video", JingleState.CONNECTION_RECEIVED, event.sid)

        if (
            self._jingle_audio_state in (JingleState.CONNECTED, JingleState.CONNECTING)
            and "video" in content_types
        ):
            self._incoming_video_event = event
        else:
            # There is no voice call running running yet
            sound.play("incoming-call-sound", event.account, loop=True)

            client = app.get_client(event.account)
            contact = client.get_module("Contacts").get_contact(event.jid)
            assert isinstance(contact, BareContact)
            app.ged.raise_event(
                events.Notification(
                    context_id="",
                    account=event.account,
                    jid=event.jid,
                    type="incoming-call",
                    title=_("Incoming Call"),
                    text=_("%s is calling") % contact.name,
                )
            )

    def _on_jingle_connected(self, event: events.JingleConnectedReceived) -> None:
        client = app.get_client(event.account)
        session = client.get_module("Jingle").get_jingle_session(event.fjid, event.sid)
        if session is None:
            return

        if event.media == "audio":
            self._set_jingle_state("audio", JingleState.CONNECTED, event.sid)
        if event.media == "video":
            self._set_jingle_state("video", JingleState.CONNECTED, event.sid)

        if not session.accepted:
            session.approve_session()
        for content in event.media:
            session.approve_content(content)
        sound.stop()

    def _on_jingle_disconnected(self, event: events.JingleDisconnectedReceived) -> None:
        if event.media is None:
            self._stop_jingle(
                event.account,
                JID.from_string(event.fjid),
                sid=event.sid,
                reason=event.reason,
            )
        if event.media == "audio":
            self._set_jingle_state(
                "audio", JingleState.NULL, sid=event.sid, reason=event.reason
            )
        if event.media == "video":
            self._set_jingle_state(
                "video", JingleState.NULL, sid=event.sid, reason=event.reason
            )
        sound.stop()

    def _on_jingle_error(self, event: events.JingleErrorReceived) -> None:
        if event.sid == self._jingle_audio_sid:
            self._set_jingle_state("audio", JingleState.ERROR, reason=event.reason)
        sound.stop()

    def _set_jingle_state(
        self,
        jingle_type: str,
        state: JingleState,
        sid: str | None = None,
        reason: str | None = None,
    ) -> None:
        if (
            state
            in (
                JingleState.CONNECTING,
                JingleState.CONNECTED,
                JingleState.NULL,
                JingleState.ERROR,
            )
            and reason
        ):
            log.info("%s state: %s, reason: %s", jingle_type, state, reason)

        if state == JingleState.ERROR:
            return

        if jingle_type == "audio":
            if state == self._jingle_audio_state:
                return
            if state == JingleState.NULL and self._jingle_audio_sid not in (None, sid):
                return
        else:
            if state == self._jingle_video_state:
                return
            if state == JingleState.NULL and self._jingle_video_sid not in (None, sid):
                return

        new_sid = None
        if state == JingleState.NULL:
            new_sid = None
        if state in (
            JingleState.CONNECTION_RECEIVED,
            JingleState.CONNECTING,
            JingleState.CONNECTED,
        ):
            new_sid = sid

        if jingle_type == "audio":
            self._jingle_audio_state = state
            self._jingle_audio_sid = new_sid
        else:
            self._jingle_video_state = state
            self._jingle_video_sid = new_sid

        app.ged.raise_event(
            events.CallUpdated(
                jingle_type=jingle_type,
                audio_state=self._jingle_audio_state,
                video_state=self._jingle_video_state,
                audio_sid=self._jingle_audio_sid,
                video_sid=self._jingle_video_sid,
            )
        )

        if (
            self._jingle_audio_state == JingleState.NULL
            and self._jingle_video_state == JingleState.NULL
        ):
            assert self._account is not None
            assert self._resource_jid is not None
            app.ged.raise_event(
                events.CallStopped(
                    self._account, JID.from_string(self._resource_jid.bare)
                )
            )

    def _stop_jingle(
        self,
        account: str,
        full_jid: JID,
        sid: str | None = None,
        reason: str | None = None,
    ) -> None:
        if self._jingle_audio_sid and sid in (self._jingle_audio_sid, None):
            self._close_jingle_content(account, full_jid, "audio")
        if self._jingle_video_sid and sid in (self._jingle_video_sid, None):
            self._close_jingle_content(account, full_jid, "video")
        if reason is not None:
            log.info("Stopping Jingle: %s", reason)

        self._account = None
        self._resource_jid = None

    def _close_jingle_content(
        self,
        account: str,
        full_jid: JID,
        jingle_type: str,
        shutdown: bool | None = False,
    ) -> None:
        if jingle_type == "audio":
            if self._jingle_audio_sid is None:
                return
        else:
            if self._jingle_video_sid is None:
                return

        client = app.get_client(account)
        if jingle_type == "audio":
            session = client.get_module("Jingle").get_jingle_session(
                str(full_jid), self._jingle_audio_sid
            )
        else:
            session = client.get_module("Jingle").get_jingle_session(
                str(full_jid), self._jingle_video_sid
            )

        if session is not None:
            content = session.get_content(jingle_type)
            if content is not None:
                assert content.creator is not None
                assert content.name is not None
                session.remove_content(content.creator, content.name)

        if not shutdown:
            if jingle_type == "audio":
                self._jingle_audio_sid = None
                self._jingle_audio_state = JingleState.NULL
            else:
                self._jingle_video_sid = None
                self._jingle_video_state = JingleState.NULL

        app.ged.raise_event(
            events.CallUpdated(
                jingle_type=jingle_type,
                audio_state=self._jingle_audio_state,
                video_state=self._jingle_video_state,
                audio_sid=self._jingle_audio_sid,
                video_sid=self._jingle_video_sid,
            )
        )

    def _get_audio_content(self, account: str, jid: JID) -> JingleAudio | None:
        client = app.get_client(account)
        session = client.get_module("Jingle").get_jingle_session(
            str(jid), self._jingle_audio_sid
        )
        if session is None:
            return None

        content = session.get_content("audio")
        if content is None:
            return None

        assert isinstance(content, JingleAudio)
        return content

    def start_call(self, account: str, jid: JID, call_type: CallType) -> None:
        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(contact, BareContact)
        call_resources = self._get_call_resources(contact)
        if not call_resources:
            return

        # TODO: Resource priority
        self._account = account
        self._resource_jid = call_resources[0].jid

        if self._jingle_audio_sid is None:
            app.ged.raise_event(events.CallStarted(self._account, self._resource_jid))

        if call_type == CallType.VIDEO:
            sid = client.get_module("Jingle").start_audio_video(str(self._resource_jid))
            self._set_jingle_state("audio", JingleState.CONNECTING, sid)
            self._set_jingle_state("video", JingleState.CONNECTING, sid)
            self._store_outgoing_call(account, jid, sid)
            return

        if self._jingle_audio_state != JingleState.NULL:
            self._close_jingle_content(account, self._resource_jid, "audio")
        else:
            sid = client.get_module("Jingle").start_audio(str(self._resource_jid))
            self._set_jingle_state("audio", JingleState.CONNECTING, sid)
            self._store_outgoing_call(account, jid, sid)

    def stop_call(self, account: str, contact: types.ResourceContact) -> None:
        self._close_jingle_content(account, contact.jid, "audio", shutdown=True)
        self._close_jingle_content(account, contact.jid, "video", shutdown=True)

        self._account = None
        self._resource_jid = None

    def upgrade_to_video_call(self) -> None:
        assert self._incoming_video_event is not None
        self.accept_call(self._incoming_video_event.jingle_session)
        self._incoming_video_event = None

    def accept_call(self, session: JingleSession) -> None:
        if self._jingle_video_sid is None:
            assert self._account is not None
            assert self._resource_jid is not None
            app.ged.raise_event(events.CallStarted(self._account, self._resource_jid))

        sound.stop()

        audio = session.get_content("audio")
        video = session.get_content("video")

        if audio and not audio.negotiated:
            self._set_jingle_state("audio", JingleState.CONNECTING, session.sid)
        if video and not video.negotiated:
            self._set_jingle_state("video", JingleState.CONNECTING, session.sid)

        if not session.accepted:
            session.approve_session()

        if audio is not None:
            session.approve_content("audio")
        if video is not None:
            session.approve_content("video")

    def get_active_call_jid(self) -> JID | None:
        return self._resource_jid

    @staticmethod
    def decline_call(session: JingleSession) -> None:
        sound.stop()

        audio = session.get_content("audio")
        video = session.get_content("video")

        if not session.accepted:
            session.decline_session()
        else:
            if audio is not None:
                session.reject_content("audio")
            if video is not None:
                session.reject_content("video")

    def start_dtmf(self, account: str, jid: JID, key: str) -> None:
        content = self._get_audio_content(account, jid)
        if content is not None:
            content.start_dtmf(key)

    def stop_dtmf(self, account: str, jid: JID) -> None:
        content = self._get_audio_content(account, jid)
        if content is not None:
            content.stop_dtmf()

    def mic_volume_changed(self, account: str, jid: JID, value: int) -> None:
        content = self._get_audio_content(account, jid)
        if content is not None:
            content.set_mic_volume(value / 100)
            app.settings.set("audio_input_volume", value)

    def output_volume_changed(self, account: str, jid: JID, value: int) -> None:
        content = self._get_audio_content(account, jid)
        if content is not None:
            content.set_out_volume(value / 100)
            app.settings.set("audio_output_volume", value)

    @staticmethod
    def _get_call_resources(contact: types.BareContact) -> list[types.ResourceContact]:
        resource_list: list[types.ResourceContact] = []
        for resource_contact in contact.iter_resources():
            if resource_contact.supports(Namespace.JINGLE_RTP):
                resource_list.append(resource_contact)
        return resource_list

    @staticmethod
    def _store_outgoing_call(account: str, jid: JID, sid: str) -> None:
        call_data = mod.Call(
            sid=sid,
            state=0,  # TODO
        )

        message = mod.Message(
            account_=account,
            remote_jid_=jid,
            resource=None,
            type=MessageType.CHAT,
            direction=ChatDirection.OUTGOING,
            timestamp=utc_now(),
            state=MessageState.ACKNOWLEDGED,
            id=str(uuid.uuid4()),
            call=call_data,
        )

        app.storage.archive.insert_object(message)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


# Handles Jingle Transports (currently only ICE-UDP)


from __future__ import annotations

from typing import Any

import logging
import socket
import uuid
from enum import IntEnum
from enum import unique

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.util import generate_id

from gajim.common import app
from gajim.common import types
from gajim.common.file_props import FileProp
from gajim.common.jingle_content import JingleContent

log = logging.getLogger("gajim.c.jingle_transport")

transports: dict[str, Any] = {}

if app.is_installed("FARSTREAM"):
    from gi.repository import Farstream


def get_jingle_transport(node: nbxmpp.Node):
    namespace = node.getNamespace()
    if namespace in transports:
        return transports[namespace](node)


@unique
class TransportType(IntEnum):
    """
    Possible types of a JingleTransport
    """

    ICEUDP = 1
    SOCKS5 = 2
    IBB = 3


class JingleTransport:
    """
    An abstraction of a transport in Jingle sessions
    """

    __slots__ = [
        "candidates",
        "connection",
        "file_props",
        "ourjid",
        "remote_candidates",
        "sid",
        "type_",
    ]

    def __init__(self, type_: TransportType) -> None:
        self.type_ = type_
        self.candidates: list[dict[str, Any]] = []
        self.remote_candidates: list[dict[str, Any]] = []

        self.connection: types.Client | None = None
        self.file_props: FileProp | None = None
        self.ourjid: str | None = None
        self.sid: str | None = None

    def _iter_candidates(self):
        for candidate in self.candidates:
            yield self.make_candidate(candidate)

    def make_candidate(self, candidate: dict[str, Any]) -> Any:
        """
        Build a candidate stanza for the given candidate
        """

    def make_transport(
        self, candidates: list[dict[str, Any]] | None = None
    ) -> nbxmpp.Node:
        """
        Build a transport stanza with the given candidates (or self.candidates
        if candidates is None)
        """
        if not candidates:
            candidates = list(self._iter_candidates())
        else:
            candidates = (self.make_candidate(candidate) for candidate in candidates)
        transport = nbxmpp.Node("transport", payload=candidates)
        return transport

    def parse_transport_stanza(self, transport: nbxmpp.Node) -> list[Any]:
        """
        Return the list of transport candidates from a transport stanza
        """
        return []

    def set_connection(self, conn: types.Client) -> None:
        self.connection = conn
        if not self.sid:
            self.sid = generate_id()

    def set_file_props(self, file_props: FileProp) -> None:
        self.file_props = file_props

    def set_our_jid(self, jid: str) -> None:
        self.ourjid = jid

    def set_sid(self, sid: str) -> None:
        self.sid = sid


class JingleTransportSocks5(JingleTransport):
    """
    Socks5 transport in jingle scenario
    Note: Don't forget to call set_file_props after initialization
    """

    def __init__(self, node: nbxmpp.Node | None = None) -> None:
        JingleTransport.__init__(self, TransportType.SOCKS5)
        self.connection = None
        self.remote_candidates: list[dict[str, Any]] = []
        self.sid: str | None = None
        if node and node.getAttr("sid"):
            self.sid = node.getAttr("sid")

    def make_candidate(self, candidate) -> nbxmpp.Node:
        log.info("candidate dict, %s", candidate)
        attrs = {
            "cid": candidate["candidate_id"],
            "host": candidate["host"],
            "jid": candidate["jid"],
            "port": candidate["port"],
            "priority": candidate["priority"],
            "type": candidate["type"],
        }

        return nbxmpp.Node("candidate", attrs=attrs)

    def make_transport(
        self,
        candidates: list[dict[str, Any]] | None = None,
        add_candidates: bool = True,
    ) -> nbxmpp.Node:
        if add_candidates:
            self._add_local_ips_as_candidates()
            self._add_additional_candidates()
            self._add_proxy_candidates()
            transport = JingleTransport.make_transport(self, candidates)
        else:
            transport = nbxmpp.Node("transport")
        transport.setNamespace(Namespace.JINGLE_BYTESTREAM)
        transport.setAttr("sid", self.sid)
        if self.file_props.dstaddr:
            transport.setAttr("dstaddr", self.file_props.dstaddr)
        return transport

    def parse_transport_stanza(self, transport: nbxmpp.Node) -> list[dict[str, Any]]:
        candidates: list[dict[str, Any]] = []
        for candidate in transport.iterTags("candidate"):
            cand = {
                "state": 0,
                "target": self.ourjid,
                "host": candidate["host"],
                "port": int(candidate["port"] or 1080),
                "candidate_id": candidate["cid"],
                "type": candidate["type"] or "direct",
                "priority": candidate["priority"],
            }
            candidates.append(cand)

            # we need this when we construct file_props on session-initiation
        if candidates:
            self.remote_candidates = candidates
        return candidates

    def _add_candidates(self, candidates: list[dict[str, Any]]) -> None:
        for cand in candidates:
            in_remote = False
            for cand2 in self.remote_candidates:
                if cand["host"] == cand2["host"] and cand["port"] == cand2["port"]:
                    in_remote = True
                    break
            if not in_remote:
                self.candidates.append(cand)

    def _add_local_ips_as_candidates(self) -> None:
        if not app.settings.get_account_setting(
            self.connection.name, "ft_send_local_ips"
        ):
            return
        if not self.connection:
            return
        port = int(app.settings.get("file_transfers_port"))
        # type preference of connection type. XEP-0260 section 2.2
        type_preference = 126
        priority = (2**16) * type_preference

        hosts: set[str | None] = set()
        local_ip_cand: list[dict[str, Any]] = []

        my_ip = self.connection.local_address
        if my_ip is None:
            log.warning("No local address available")

        else:
            candidate = {
                "host": my_ip,
                "candidate_id": generate_id(),
                "port": port,
                "type": "direct",
                "jid": self.ourjid,
                "priority": priority,
            }
            hosts.add(my_ip)
            local_ip_cand.append(candidate)

        try:
            for addrinfo in socket.getaddrinfo(socket.gethostname(), None):
                addr = addrinfo[4][0]
                if addr not in hosts and not addr.startswith("127.") and addr != "::1":
                    candidate = {
                        "host": addr,
                        "candidate_id": generate_id(),
                        "port": port,
                        "type": "direct",
                        "jid": self.ourjid,
                        "priority": priority,
                        "initiator": self.file_props.sender,
                        "target": self.file_props.receiver,
                    }
                    hosts.add(addr)
                    local_ip_cand.append(candidate)
        except socket.gaierror:
            pass  # ignore address-related errors for getaddrinfo

        try:
            from netifaces import AF_INET
            from netifaces import AF_INET6
            from netifaces import ifaddresses
            from netifaces import interfaces

            for ifaceName in interfaces():
                addresses = ifaddresses(ifaceName)
                if AF_INET in addresses:
                    for address in addresses[AF_INET]:
                        addr = address["addr"]
                        if addr in hosts or addr.startswith("127."):
                            continue
                        candidate = {
                            "host": addr,
                            "candidate_id": generate_id(),
                            "port": port,
                            "type": "direct",
                            "jid": self.ourjid,
                            "priority": priority,
                            "initiator": self.file_props.sender,
                            "target": self.file_props.receiver,
                        }
                        hosts.add(addr)
                        local_ip_cand.append(candidate)
                if AF_INET6 in addresses:
                    for address in addresses[AF_INET6]:
                        addr = address["addr"]
                        if (
                            addr in hosts
                            or addr.startswith("::1")
                            or addr.count(":") != 7
                        ):
                            continue
                        candidate = {
                            "host": addr,
                            "candidate_id": generate_id(),
                            "port": port,
                            "type": "direct",
                            "jid": self.ourjid,
                            "priority": priority,
                            "initiator": self.file_props.sender,
                            "target": self.file_props.receiver,
                        }
                        hosts.add(addr)
                        local_ip_cand.append(candidate)

        except ImportError:
            pass

        self._add_candidates(local_ip_cand)

    def _add_additional_candidates(self) -> None:
        if not self.connection:
            return
        type_preference = 126
        priority = (2**16) * type_preference
        additional_ip_cand = []
        port = int(app.settings.get("file_transfers_port"))
        ft_add_hosts = app.settings.get("ft_add_hosts_to_send")

        if ft_add_hosts:
            hosts = [e.strip() for e in ft_add_hosts.split(",")]
            for host in hosts:
                candidate = {
                    "host": host,
                    "candidate_id": generate_id(),
                    "port": port,
                    "type": "direct",
                    "jid": self.ourjid,
                    "priority": priority,
                    "initiator": self.file_props.sender,
                    "target": self.file_props.receiver,
                }
                additional_ip_cand.append(candidate)

        self._add_candidates(additional_ip_cand)

    def _add_proxy_candidates(self) -> None:
        if not self.connection:
            return
        type_preference = 10
        priority = (2**16) * type_preference
        proxy_cand = []
        socks5conn = self.connection
        proxyhosts = socks5conn.get_module(
            "Bytestream"
        )._get_file_transfer_proxies_from_config(self.file_props)

        if proxyhosts:
            self.file_props.proxyhosts = proxyhosts

            for proxyhost in proxyhosts:
                candidate = {
                    "host": proxyhost["host"],
                    "candidate_id": generate_id(),
                    "port": int(proxyhost["port"]),
                    "type": "proxy",
                    "jid": proxyhost["jid"],
                    "priority": priority,
                    "initiator": self.file_props.sender,
                    "target": self.file_props.receiver,
                }
                proxy_cand.append(candidate)

        self._add_candidates(proxy_cand)

    def get_content(self) -> JingleContent | None:
        sesn = self.connection.get_module("Jingle").get_jingle_session(
            self.ourjid, self.file_props.sid
        )
        for content in sesn.contents.values():
            if content.transport == self:
                return content
        return None

    def _on_proxy_auth_ok(self, proxy):
        log.info("proxy auth ok for %s", str(proxy))
        # send activate request to proxy, send activated confirmation to peer
        if not self.connection:
            return
        sesn = self.connection.get_module("Jingle").get_jingle_session(
            self.ourjid, self.file_props.sid
        )
        if sesn is None:
            return

        iq = nbxmpp.Iq(to=proxy["jid"], frm=self.ourjid, typ="set")
        auth_id = "au_" + proxy["sid"]
        iq.setID(auth_id)
        query = iq.setTag("query", namespace=Namespace.BYTESTREAM)
        query.setAttr("sid", proxy["sid"])
        activate = query.setTag("activate")
        activate.setData(sesn.peerjid)
        iq.setID(auth_id)
        self.connection.connection.send(iq)

        content = nbxmpp.Node("content")
        content.setAttr("creator", "initiator")
        content_object = self.get_content()
        content.setAttr("name", content_object.name)
        transport = nbxmpp.Node("transport")
        transport.setNamespace(Namespace.JINGLE_BYTESTREAM)
        transport.setAttr("sid", proxy["sid"])
        activated = nbxmpp.Node("activated")
        cid = None

        if "cid" in proxy:
            cid = proxy["cid"]
        else:
            for host in self.candidates:
                if (
                    host["host"] == proxy["host"]
                    and host["jid"] == proxy["jid"]
                    and host["port"] == proxy["port"]
                ):
                    cid = host["candidate_id"]
                    break
        if cid is None:
            raise Exception("cid is missing")
        activated.setAttr("cid", cid)
        transport.addChild(node=activated)
        content.addChild(node=transport)
        sesn.send_transport_info(content)


class JingleTransportIBB(JingleTransport):
    def __init__(
        self, node: nbxmpp.Node | None = None, block_sz: str | None = None
    ) -> None:

        JingleTransport.__init__(self, TransportType.IBB)

        if block_sz:
            self.block_sz = block_sz
        else:
            self.block_sz = "4096"

        self.connection = None
        self.sid: str | None = None
        if node and node.getAttr("sid"):
            self.sid = node.getAttr("sid")

    def make_transport(self) -> nbxmpp.Node:
        transport = nbxmpp.Node("transport")
        transport.setNamespace(Namespace.JINGLE_IBB)
        transport.setAttr("block-size", self.block_sz)
        transport.setAttr("sid", self.sid)
        return transport


class JingleTransportICEUDP(JingleTransport):
    def __init__(self, node: nbxmpp.Node | None) -> None:
        JingleTransport.__init__(self, TransportType.ICEUDP)

    def make_candidate(self, candidate) -> nbxmpp.Node:
        types = {
            Farstream.CandidateType.HOST: "host",
            Farstream.CandidateType.SRFLX: "srflx",
            Farstream.CandidateType.PRFLX: "prflx",
            Farstream.CandidateType.RELAY: "relay",
            Farstream.CandidateType.MULTICAST: "multicast",
        }
        attrs: dict[str, Any] = {
            "component": candidate.component_id,
            "foundation": "1",  # hack
            "generation": "0",
            "ip": candidate.ip,
            "network": "0",
            "port": candidate.port,
            "priority": int(candidate.priority),  # hack
            "id": str(uuid.uuid4()),
        }
        if candidate.type in types:
            attrs["type"] = types[candidate.type]
        if candidate.proto == Farstream.NetworkProtocol.UDP:
            attrs["protocol"] = "udp"
        else:
            # we actually don't handle properly different tcp options in jingle
            attrs["protocol"] = "tcp"
        return nbxmpp.Node("candidate", attrs=attrs)

    def make_transport(
        self, candidates: list[dict[str, Any]] | None = None
    ) -> nbxmpp.Node:
        transport = JingleTransport.make_transport(self, candidates)
        transport.setNamespace(Namespace.JINGLE_ICE_UDP)
        if (
            self.candidates
            and self.candidates[0].username
            and self.candidates[0].password
        ):
            transport.setAttr("ufrag", self.candidates[0].username)
            transport.setAttr("pwd", self.candidates[0].password)
        return transport

    def parse_transport_stanza(
        self, transport: nbxmpp.Node
    ) -> list[Farstream.Candidate]:
        candidates: list[Farstream.Candidate] = []
        for candidate in transport.iterTags("candidate"):
            foundation = str(candidate["foundation"])
            component_id = int(candidate["component"])
            ip = str(candidate["ip"])
            port = int(candidate["port"])
            base_ip = None
            base_port = 0
            if candidate["protocol"] == "udp":
                proto = Farstream.NetworkProtocol.UDP
            else:
                # we actually don't handle properly different tcp options in
                # jingle
                proto = Farstream.NetworkProtocol.TCP
            priority = int(candidate["priority"])
            types = {
                "host": Farstream.CandidateType.HOST,
                "srflx": Farstream.CandidateType.SRFLX,
                "prflx": Farstream.CandidateType.PRFLX,
                "relay": Farstream.CandidateType.RELAY,
                "multicast": Farstream.CandidateType.MULTICAST,
            }
            if "type" in candidate and candidate["type"] in types:
                type_ = types[candidate["type"]]
            else:
                log.warning("Unknown type %s", candidate["type"])
                type_ = Farstream.CandidateType.HOST
            username = str(transport["ufrag"])
            password = str(transport["pwd"])
            ttl = 0

            cand = Farstream.Candidate.new_full(
                foundation,
                component_id,
                ip,
                port,
                base_ip,
                base_port,
                proto,
                priority,
                type_,
                username,
                password,
                ttl,
            )

            candidates.append(cand)
        self.remote_candidates.extend(candidates)
        return candidates


transports[Namespace.JINGLE_ICE_UDP] = JingleTransportICEUDP
transports[Namespace.JINGLE_BYTESTREAM] = JingleTransportSocks5
transports[Namespace.JINGLE_IBB] = JingleTransportIBB



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast

import json
import logging
from collections.abc import Iterable
from io import BytesIO
from zipfile import ZipFile

from gi.repository import GLib

from gajim.common import app
from gajim.common import configpaths
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.helpers import Observable

from .manifest import PluginManifest

log = logging.getLogger("gajim.p.repository")


REPO_INDEX_FILENAME = "package_index.json"
UPDATE_CHECK_INTERVAL = 68400


class PluginRepository(Observable):
    """Signals

    - donwload-started
    - download-finished
    - download-failed
    - plugin-updates-available
    - auto-update-finished
    """

    def __init__(self) -> None:
        Observable.__init__(self)
        self._manifests: dict[str, PluginManifest] = {}

        self._repository_url = ""
        self._repository_index_url = ""
        self._download_path = configpaths.get("PLUGINS_DOWNLOAD")

        self._auto_update_in_progress = False

        self._download_queue: set[PluginManifest] = set()

        if app.settings.get("plugins_repository_enabled"):
            app.ftm.http_request(
                "GET",
                "https://gajim.org/updates.json",
                callback=self._on_repository_received,
            )

    @property
    def available(self):
        return bool(self._repository_url)

    def _on_repository_received(self, obj: FileTransfer) -> None:
        try:
            result = obj.get_result()
        except Exception as error:
            log.warning("Repository retrieval failed: %s", error)
            return

        try:
            updates = json.loads(result.content)
        except Exception as error:
            log.warning("Unable to parse repository information: %s", error)
            return

        repository_url = updates.get("plugin_repository")
        if repository_url is None:
            log.warning("No plugin repository url found")
            return

        self.set_repository(repository_url)

    def get_manifests(self) -> list[PluginManifest]:
        return sorted(self._manifests.values(), key=lambda m: m.name)

    def contains(self, short_name: str) -> bool:
        return self._manifests.get(short_name) is not None

    def set_repository(self, repository_url: str) -> None:
        self._repository_url = repository_url
        self._repository_index_url = f"{repository_url}/{REPO_INDEX_FILENAME}"
        self._refresh_plugin_index()
        if app.settings.get("plugins_update_check"):
            GLib.timeout_add_seconds(30, self._check_if_updates_are_needed)

    def _refresh_update_check_timer(self, seconds: int) -> None:
        GLib.timeout_add_seconds(
            seconds, self._refresh_plugin_index, self._check_if_updates_are_needed
        )

    def _plugin_needs_update(self, manifest: PluginManifest) -> bool:
        installed_plugin = app.plugin_manager.get_plugin(manifest.short_name)
        if installed_plugin is None:
            return False

        path = configpaths.get("PLUGINS_DOWNLOAD") / manifest.short_name
        if path.exists():
            return False

        return installed_plugin.manifest.version < manifest.version

    def _parse_package_json(
        self, package_index: dict[str, Any]
    ) -> dict[str, PluginManifest]:
        suitable_plugins: dict[str, PluginManifest] = {}
        plugins = package_index["plugins"]
        for name, versions in plugins.items():
            for version, manifest in versions.items():
                manifest["short_name"] = name
                manifest["version"] = version
                try:
                    manifest = PluginManifest.from_manifest_json(manifest)
                except Exception as error:
                    log.info("Invalid manifest: %s", error)
                    continue

                if not manifest.is_usable:
                    log.info(
                        "Ignore not suitable plugin: %s %s",
                        manifest.short_name,
                        manifest.version,
                    )
                    continue

                last_plugin = suitable_plugins.get(manifest.short_name)
                if last_plugin is None or last_plugin.version < manifest.version:
                    suitable_plugins[manifest.short_name] = manifest

        return suitable_plugins

    def _refresh_plugin_index(self, callback: Any | None = None) -> None:
        log.info("Refresh index")
        app.ftm.http_request(
            "GET",
            self._repository_index_url,
            user_data=callback,
            callback=self._on_index_received,
        )

    def _on_index_received(self, obj: FileTransfer) -> None:
        try:
            result = obj.get_result()
        except Exception as error:
            log.warning("Refresh failed: %s %s", self._repository_index_url, error)
            return

        try:
            package_index = json.loads(result.content)
        except Exception as error:
            log.warning("Unable to parse repository index: %s", error)
            return

        self._manifests = self._parse_package_json(package_index)

        image_path = package_index["metadata"]["image_path"]
        callback = obj.get_user_data()

        app.ftm.http_request(
            "GET",
            f"{self._repository_url}/{image_path}",
            callback=self._on_images_received,
        )

        log.info("Refresh successful")

        if callback is not None:
            callback()

    def _on_images_received(self, obj: FileTransfer) -> None:
        try:
            result = obj.get_result()
        except Exception as error:
            log.warning("Image download failed: %s", error)
            return

        path = configpaths.get("PLUGINS_IMAGES")

        try:
            zipfile = ZipFile(BytesIO(result.content))
            zipfile.extractall(path)
        except Exception as error:
            log.warning("Failed to extract images: %s", error)

    def _check_if_updates_are_needed(self) -> None:
        plugins_to_update: list[PluginManifest] = []
        for manifest in self._manifests.values():
            if not self._plugin_needs_update(manifest):
                continue

            log.info(
                "Update available for: %s - %s", manifest.short_name, manifest.version
            )
            plugins_to_update.append(manifest)

        if not plugins_to_update:
            log.info("No updates available")
            return

        if app.settings.get("plugins_auto_update"):
            self._auto_update_in_progress = True
            self.download_plugins(plugins_to_update)
        else:
            self.notify("plugin-updates-available", plugins_to_update)

        self._refresh_update_check_timer(UPDATE_CHECK_INTERVAL)

    def download_plugins(self, manifests: Iterable[PluginManifest]) -> None:
        if not self._repository_url:
            log.warning("Unable to download plugins because repository not set")
            return

        manifests = [self._manifests[manifest.short_name] for manifest in manifests]

        manifests = set(manifests) - self._download_queue
        if not manifests:
            return

        self._download_queue |= manifests
        self.notify("download-started", manifests)

        for manifest in manifests:
            self._download_plugin(manifest)

    def _download_plugin(self, manifest: PluginManifest) -> None:
        log.info("Download plugin %s", manifest.short_name)
        url = manifest.get_remote_url(self._repository_url)

        app.ftm.http_request(
            "GET",
            url,
            user_data=manifest,
            callback=self._on_download_plugin_finished,
        )

    def _on_download_plugin_finished(self, obj: FileTransfer) -> None:
        manifest = cast(PluginManifest, obj.get_user_data())
        self._download_queue.remove(manifest)

        try:
            result = obj.get_result()
        except Exception as error:
            log.error(
                "Download failed: %s %s",
                manifest.get_remote_url(self._repository_url),
                error,
            )
            self.notify("download-failed", manifest, str(error))
            return

        with ZipFile(BytesIO(result.content)) as zip_file:
            zip_file.extractall(self._download_path / manifest.short_name)

        log.info("Finished downloading %s", manifest.short_name)
        self.notify("download-finished", manifest)

        if len(self._download_queue) == 0:
            if self._auto_update_in_progress:
                self._auto_update_in_progress = False
                if app.settings.get("plugins_notify_after_update"):
                    self.notify("auto-update-finished")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0048: Bookmarks

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal

from gi.repository import GLib
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import BookmarkData
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MessageProperties
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.events import BookmarksReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.util import event_node

NODE_MAX_NS = "http://jabber.org/protocol/pubsub#config-node-max"


class Bookmarks(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._bookmark_event_received)
        self._register_pubsub_handler(self._bookmark_1_event_received)
        self._conversion = False
        self._compat = False
        self._compat_pep = False
        self._node_max = False
        self._bookmarks: types.BookmarksDict = {}
        self._join_timeouts: list[int] = []
        self._request_in_progress = True

    @property
    def conversion(self) -> bool:
        return self._conversion

    @property
    def compat(self) -> bool:
        return self._compat

    @property
    def compat_pep(self) -> bool:
        return self._compat_pep

    @property
    def bookmarks(self) -> list[BookmarkData]:
        return list(self._bookmarks.values())

    @property
    def pep_bookmarks_used(self) -> bool:
        return self._bookmark_module() == "PEPBookmarks"

    @property
    def nativ_bookmarks_used(self) -> bool:
        return self._bookmark_module() == "NativeBookmarks"

    @event_node(Namespace.BOOKMARKS)
    def _bookmark_event_received(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:
        assert properties.pubsub_event is not None
        if properties.pubsub_event.retracted:
            return

        if not properties.is_self_message:
            self._log.warning("%s has an open access bookmarks node", properties.jid)
            return

        if not self.pep_bookmarks_used:
            return

        if self._request_in_progress:
            self._log.info("Ignore update, pubsub request in progress")
            return

        bookmarks = self._convert_to_dict(properties.pubsub_event.data)

        old_bookmarks = self._bookmarks.copy()
        self._bookmarks = bookmarks
        self._act_on_changed_bookmarks(old_bookmarks)
        app.ged.raise_event(BookmarksReceived(account=self._account))

    @event_node(Namespace.BOOKMARKS_1)
    def _bookmark_1_event_received(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:
        if not properties.is_self_message:
            self._log.warning("%s has an open access bookmarks node", properties.jid)
            return

        if not self.nativ_bookmarks_used:
            return

        if self._request_in_progress:
            self._log.info("Ignore update, pubsub request in progress")
            return

        old_bookmarks = self._bookmarks.copy()

        assert properties.pubsub_event is not None
        if properties.pubsub_event.deleted or properties.pubsub_event.purged:
            self._log.info("Bookmark node deleted/purged")
            self._bookmarks.clear()

        elif properties.pubsub_event.retracted:
            try:
                jid = JID.from_string(properties.pubsub_event.id)
            except Exception as error:
                self._log.info(
                    "Unable to parse retracted bookmark: %s %s",
                    properties.pubsub_event.id,
                    error,
                )
            else:
                self._log.info("Retract: %s", jid)
                bookmark = self._bookmarks.get(jid)
                if bookmark is not None:
                    self._bookmarks.pop(jid, None)

        else:
            new_bookmark = cast(BookmarkData, properties.pubsub_event.data)
            self._bookmarks[new_bookmark.jid] = new_bookmark

        self._act_on_changed_bookmarks(old_bookmarks)
        app.ged.raise_event(BookmarksReceived(account=self._account))

    def pass_disco(self, info: DiscoInfo) -> None:
        self._node_max = NODE_MAX_NS in info.features
        self._compat_pep = Namespace.BOOKMARKS_COMPAT_PEP in info.features
        self._compat = Namespace.BOOKMARKS_COMPAT in info.features
        self._conversion = Namespace.BOOKMARK_CONVERSION in info.features

    def _bookmark_module(
        self,
    ) -> Literal["PrivateBookmarks", "NativeBookmarks", "PEPBookmarks"]:
        if not self._con.get_module("PubSub").publish_options:
            return "PrivateBookmarks"

        if app.settings.get("dev_force_bookmark_2"):
            return "NativeBookmarks"

        if self._compat_pep and self._node_max:
            return "NativeBookmarks"

        if self._conversion:
            return "PEPBookmarks"
        return "PrivateBookmarks"

    def _act_on_changed_bookmarks(self, current_bookmarks: types.BookmarksDict) -> None:

        changed_autojoin, changed_name = self._get_changed_attrs(
            current_bookmarks, self._bookmarks
        )

        if changed_name:
            app.storage.archive.bulk_update_custom_names(self._account, changed_name)

        join_bookmarks = [
            self._bookmarks[jid] for jid, autojoin in changed_autojoin if autojoin
        ]
        self._log.info(
            "Schedule autojoin in 10s for: %s\n", "\n".join(str(join_bookmarks))
        )

        # If another client creates a MUC, the MUC is locked until the
        # configuration is finished. Give the user some time to finish
        # the configuration.
        if app.settings.get_account_setting(self._account, "autojoin_sync"):
            timeout_id = GLib.timeout_add_seconds(
                10, self._join_with_timeout, join_bookmarks
            )
            self._join_timeouts.append(timeout_id)

        # TODO: leave mucs
        # leave = [jid for jid, autojoin in changed if not autojoin]

    @staticmethod
    def _get_changed_attrs(
        old_bookmarks: types.BookmarksDict,
        new_bookmarks: types.BookmarksDict,
    ) -> tuple[set[tuple[JID, bool]], set[tuple[JID, str | None]]]:

        new_autojoin: set[tuple[JID, bool]] = set()
        new_name: set[tuple[JID, str | None]] = set()
        for b in new_bookmarks.values():
            new_autojoin.add((b.jid, b.autojoin))
            new_name.add((b.jid, b.name))

        old_autojoin: set[tuple[JID, bool]] = set()
        old_name: set[tuple[JID, str | None]] = set()
        for b in old_bookmarks.values():
            old_autojoin.add((b.jid, b.autojoin))
            old_name.add((b.jid, b.name))

        changed_autojoin = new_autojoin - old_autojoin
        changed_name = new_name - old_name
        return changed_autojoin, changed_name

    @staticmethod
    def _convert_to_dict(bookmarks: list[BookmarkData] | None) -> types.BookmarksDict:
        _dict: types.BookmarksDict = {}
        if not bookmarks:
            return _dict

        for bookmark in bookmarks:
            _dict[bookmark.jid] = bookmark
        return _dict

    def get_bookmark(self, jid: str | JID) -> BookmarkData | None:
        return self._bookmarks.get(cast(JID, jid))

    def request_bookmarks(self) -> None:
        if not app.account_is_available(self._account):
            return

        self._request_in_progress = True
        self._nbxmpp(self._bookmark_module()).request_bookmarks(
            callback=self._bookmarks_received
        )

    def _bookmarks_received(self, task: Task) -> None:
        try:
            bookmarks: list[BookmarkData] = task.finish()
        except Exception as error:
            self._log.warning(error)
            bookmarks = []

        self._request_in_progress = False

        self._client.get_module("MucBlocking").pass_bookmarks(bookmarks)

        self._cleanup_bookmarks(bookmarks)
        self._bookmarks = self._convert_to_dict(bookmarks)
        if app.settings.get_account_setting(self._account, "autojoin_sync"):
            self.auto_join_bookmarks(self.bookmarks)
        app.ged.raise_event(BookmarksReceived(account=self._account))

    def _cleanup_bookmarks(self, bookmarks: list[BookmarkData]) -> None:
        for bookmark in list(bookmarks):
            contact = self._client.get_module("Contacts").get_contact(
                bookmark.jid, groupchat=True
            )
            if not isinstance(contact, GroupchatContact):
                # The contact exists probably in the roster and is therefore
                # assumed to not be a MUC
                self._log.warning(
                    "Received bookmark but jid is not a groupchat: %s", bookmark.jid
                )
                bookmarks.remove(bookmark)

    def store_bookmarks(self, bookmarks: list[BookmarkData]) -> None:
        if not app.account_is_available(self._account):
            return

        if not self.nativ_bookmarks_used:
            bookmarks = self.bookmarks

        self._nbxmpp(self._bookmark_module()).store_bookmarks(bookmarks)

        app.ged.raise_event(BookmarksReceived(account=self._account))

    def _join_with_timeout(self, bookmarks: list[BookmarkData]) -> None:
        self._join_timeouts.pop(0)
        self.auto_join_bookmarks(bookmarks)

    def auto_join_bookmarks(self, bookmarks: list[BookmarkData]) -> None:
        for bookmark in bookmarks:
            if bookmark.autojoin:
                self._log.info("Autojoin Bookmark: %s", bookmark.jid)
                self._con.get_module("MUC").join(bookmark.jid)

    def modify(self, jid: JID, **kwargs: Any) -> None:
        bookmark = self._bookmarks.get(jid)
        if bookmark is None:
            return

        if "name" in kwargs:
            app.storage.archive.set_contact_value(
                self._account, jid, "custom_name", kwargs["name"]
            )

        new_bookmark = bookmark._replace(**kwargs)
        if new_bookmark == bookmark:
            # No change happened
            return
        self._log.info("Modify bookmark: %s %s", jid, kwargs)
        self._bookmarks[jid] = new_bookmark

        self.store_bookmarks([new_bookmark])

    def add_or_modify(self, jid: JID, **kwargs: Any) -> None:
        if not app.settings.get_account_setting(self._account, "autojoin_sync"):
            kwargs.pop("autojoin", None)
        bookmark = self._bookmarks.get(jid)
        if bookmark is not None:
            self.modify(jid, **kwargs)
            return

        new_bookmark = BookmarkData(jid=jid, **kwargs)
        self._bookmarks[jid] = new_bookmark
        self._log.info("Add new bookmark: %s", new_bookmark)

        app.storage.archive.set_contact_value(
            self._account, jid, "custom_name", new_bookmark.name
        )

        self.store_bookmarks([new_bookmark])

    def remove(self, jid: JID, publish: bool = True) -> None:
        removed = self._bookmarks.pop(jid, False)
        if not removed:
            return
        if publish:
            if self.nativ_bookmarks_used:
                self._nbxmpp("NativeBookmarks").retract_bookmark(jid)
            else:
                self.store_bookmarks(self.bookmarks)

    def get_name_from_bookmark(self, jid: str | JID) -> str | None:
        bookmark = self._bookmarks.get(cast(JID, jid))
        if bookmark is None:
            return bookmark
        return bookmark.name

    def is_bookmark(self, jid: str | JID) -> bool:
        return jid in self._bookmarks

    def _remove_timeouts(self) -> None:
        for _id in self._join_timeouts:
            GLib.source_remove(_id)

    def cleanup(self) -> None:
        BaseModule.cleanup(self)
        self._remove_timeouts()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import logging

from gi.repository import Gtk

from gajim.common import app
from gajim.common import passwords
from gajim.common.events import PasswordRequired
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.password_dialog")


class PasswordDialog(GajimAppWindow):
    def __init__(self, account: str, event: PasswordRequired) -> None:
        GajimAppWindow.__init__(
            self,
            name="PasswordDialog",
            title=_("Password Required"),
            default_width=400,
            add_window_padding=True,
            header_bar=True,
        )

        self._ui = get_builder("password_dialog.ui")
        self.set_child(self._ui.pass_box)

        self._connect(self._ui.cancel_button, "clicked", self._on_cancel)
        self._connect(self._ui.ok_button, "clicked", self._on_ok)

        self.account = account
        self._client = app.get_client(event.client.account)
        self._event = event

        self.set_default_widget(self._ui.ok_button)

        self._process_event()

    def _cleanup(self) -> None:
        pass

    def _process_event(self) -> None:
        own_jid = self._client.get_own_jid().bare
        account_label = app.get_account_label(self.account)

        self._ui.header.set_text(_("Password Required"))
        self._ui.message_label.set_text(
            _("Please enter your password for\n%(jid)s\n(Account: %(account)s)")
            % {"jid": own_jid, "account": account_label}
        )
        self._ui.save_pass_checkbutton.set_visible(True)

        is_keyring_available = passwords.is_keyring_available()
        self._ui.save_pass_checkbutton.set_sensitive(
            not app.settings.get("use_keyring") or is_keyring_available
        )
        if not is_keyring_available:
            self._ui.keyring_hint.set_visible(True)

    def _on_ok(self, _button: Gtk.Button) -> None:
        password = self._ui.pass_entry.get_text()
        savepass = self._ui.save_pass_checkbutton.get_active()

        app.settings.set_account_setting(self.account, "savepass", savepass)
        passwords.save_password(self.account, password)

        self._event.on_password()
        self.close()

    def _on_cancel(self, _button: Gtk.Button) -> None:
        self.close()



# Copyright (C) 2007 Jean-Marie Traissard <jim AT lapin.org>
#                    Julien Pivotto <roidelapluie AT gmail.com>
#                    Stefan Bethge <stefan AT lanpartei.de>
#                    Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2007-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import NamedTuple

import sys

from gi.repository import Adw
from gi.repository import Gtk

from gajim.common import app
from gajim.common import passwords
from gajim.common.i18n import _

from gajim.gtk.window import GajimAppWindow


class Feature(NamedTuple):
    name: str
    available: bool
    tooltip: str
    dependency_u: str
    dependency_w: str
    enabled: bool | None


class Features(GajimAppWindow):
    def __init__(self) -> None:
        GajimAppWindow.__init__(
            self,
            name="Features",
            title=_("Features"),
            add_window_padding=True,
            header_bar=True,
        )

        self.feature_listbox = Gtk.ListBox(
            selection_mode=Gtk.SelectionMode.NONE, hexpand=True
        )
        self.feature_listbox.add_css_class("boxed-list")

        preferences_group = Adw.PreferencesGroup(
            title=_("Available Features"),
            description=_("Check which features are available"),
        )
        preferences_group.add(self.feature_listbox)

        clamp = Adw.Clamp()
        clamp.set_child(preferences_group)
        self.set_child(clamp)

        for feature in self._get_features():
            self._add_feature(feature)

    def _cleanup(self) -> None:
        pass

    def _add_feature(self, feature: Feature) -> None:
        item = FeatureItem(feature)
        self.feature_listbox.append(item)

    def _get_features(self) -> list[Feature]:
        # av_available = app.is_installed("AV") and sys.platform != "win32"
        spell_check_enabled: bool = app.settings.get("use_speller")

        auto_status = [app.settings.get("autoaway"), app.settings.get("autoxa")]
        auto_status_enabled = bool(any(auto_status))

        return [
            # Feature(
            #     _("Audio / Video Calls"),
            #     av_available,
            #     _("Enables Gajim to provide Audio and Video chats"),
            #     _(
            #         "Requires: farstream-0.2, gstreamer-1.0, "
            #         "gst-plugins-base-1.0, gst-plugins-ugly-1.0, "
            #         "gst-libav"
            #     ),
            #     _("Feature not available on Windows"),
            #     None,
            # ),
            Feature(
                _("Automatic Status"),
                self._idle_available(),
                _(
                    "Enables Gajim to measure your computer’s idle time in "
                    "order to set your Status automatically"
                ),
                _(
                    "Requires:\norg.freedesktop.ScreenSaver or\norg.gnome.Mutter.IdleMonitor or\nlibxss"
                ),
                _("No additional requirements"),
                auto_status_enabled,
            ),
            Feature(
                _("Secure Password Storage"),
                passwords.is_keyring_available(),
                _(
                    "Enables Gajim to store Passwords securely instead of "
                    "storing them in plaintext"
                ),
                _("Requires: gnome-keyring or kwallet"),
                _("No additional requirements"),
                app.settings.get("use_keyring"),
            ),
            Feature(
                _("Spell Checker"),
                app.is_installed("SPELLING"),
                _("Enables Gajim to spell check your messages while composing"),
                _("Requires: libspelling"),
                _("No additional requirements"),
                spell_check_enabled,
            ),
            # Feature(
            #     _("UPnP-IGD Port Forwarding"),
            #     app.is_installed("UPNP"),
            #     _(
            #         "Enables Gajim to request your router to forward ports "
            #         "for file transfers"
            #     ),
            #     _("Requires: gupnpigd-1.0"),
            #     _("Feature not available on Windows"),
            #     None,
            # ),
        ]

    @staticmethod
    def _idle_available() -> bool:
        from gajim.common import idle

        return idle.Monitor.is_available()


class FeatureItem(Adw.ActionRow):
    def __init__(self, feature: Feature) -> None:
        Adw.ActionRow.__init__(
            self,
            title=feature.name,
            tooltip_text=feature.tooltip,
            subtitle_selectable=True,
        )

        self._icon = Gtk.Image()
        self.add_prefix(self._icon)

        if not feature.available:
            if sys.platform == "win32":
                self.set_subtitle(feature.dependency_w)
            else:
                self.set_subtitle(feature.dependency_u)

        self._set_feature(feature.available, feature.enabled)

    def do_unroot(self) -> None:
        Adw.ActionRow.do_unroot(self)
        app.check_finalize(self)

    def _set_feature(self, available: bool, enabled: bool | None) -> None:
        self._icon.remove_css_class("error")
        self._icon.remove_css_class("warning")
        self._icon.remove_css_class("success")

        if not available:
            self._icon.set_from_icon_name("lucide-x-symbolic")
            self._icon.add_css_class("error")
            return

        if enabled is not None and not enabled:
            self._icon.set_from_icon_name("lucide-circle-alert-symbolic")
            self._icon.add_css_class("warning")
            self.set_subtitle(_("Disabled in Preferences"))
        else:
            self._icon.set_from_icon_name("lucide-check-symbolic")
            self._icon.add_css_class("success")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import NamedTuple

import json
import logging
import sqlite3
import time
from collections import namedtuple

from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import RosterItem

from gajim.common import configpaths
from gajim.common.storage.base import Encoder
from gajim.common.storage.base import json_decoder
from gajim.common.storage.base import SqliteStorage
from gajim.common.storage.base import timeit

ContactCacheDictT = dict[tuple[str, JID], dict[str, Any]]

CURRENT_USER_VERSION = 10

CACHE_SQL_STATEMENT = (
    """
    CREATE TABLE caps_cache (
            hash_method TEXT,
            hash TEXT,
            data TEXT,
            last_seen INTEGER
    );
    CREATE TABLE last_seen_disco_info(
            jid TEXT PRIMARY KEY UNIQUE,
            disco_info TEXT,
            last_seen INTEGER
    );
    CREATE TABLE roster(
            account TEXT PRIMARY KEY UNIQUE,
            roster TEXT
    );
    CREATE TABLE unread(
            account TEXT,
            jid TEXT,
            count INTEGER,
            message_id TEXT,
            timestamp INTEGER,
            PRIMARY KEY (account, jid)
    );

    CREATE INDEX idx_unread ON unread(account, jid);

    PRAGMA user_version=%s;
    """
    % CURRENT_USER_VERSION
)

log = logging.getLogger("gajim.c.storage.cache")


class UnreadTableRow(NamedTuple):
    account: str
    jid: JID
    count: int  # type: ignore
    message_id: str
    timestamp: float


class CacheStorage(SqliteStorage):
    def __init__(self, in_memory: bool = False):
        path = None if in_memory else configpaths.get("CACHE_DB")
        SqliteStorage.__init__(self, log, path, CACHE_SQL_STATEMENT)

        self._entity_caps_cache: dict[tuple[str, str], DiscoInfo] = {}
        self._disco_info_cache: dict[JID, DiscoInfo] = {}

    def init(self, **kwargs: Any) -> None:
        SqliteStorage.init(self, detect_types=sqlite3.PARSE_COLNAMES)
        self._set_journal_mode("WAL")
        self._con.row_factory = self._namedtuple_factory

        self._fill_disco_info_cache()
        self._clean_caps_table()
        self._load_caps_data()

    @staticmethod
    def _namedtuple_factory(cursor: sqlite3.Cursor, row: tuple[Any, ...]) -> NamedTuple:
        assert cursor.description is not None
        fields = [col[0] for col in cursor.description]
        Row = namedtuple("Row", fields)  # type: ignore
        return Row(*row)

    def _migrate(self) -> None:
        try:
            user_version = self.user_version
        except sqlite3.DatabaseError as error:
            log.error("Database error: %s", error)
            self._reinit_storage()
            return

        if user_version > CURRENT_USER_VERSION:
            # Gajim was downgraded, reinit the storage
            self._reinit_storage()
            return

        if user_version < 10:
            self._reinit_storage()
            return

    @timeit
    def _load_caps_data(self) -> None:
        rows = self._con.execute(
            'SELECT hash_method, hash, data as "data [disco_info]" FROM caps_cache'
        )

        for row in rows:
            self._entity_caps_cache[(row.hash_method, row.hash)] = row.data

    @timeit
    def add_caps_entry(
        self, jid: JID, hash_method: str, hash_: str, caps_data: DiscoInfo
    ) -> None:
        self._entity_caps_cache[(hash_method, hash_)] = caps_data

        self._disco_info_cache[jid] = caps_data

        self._con.execute(
            """
            INSERT INTO caps_cache (hash_method, hash, data, last_seen)
            VALUES (?, ?, ?, ?)
            """,
            (hash_method, hash_, caps_data, int(time.time())),
        )
        self._delayed_commit()

    def get_caps_entry(self, hash_method: str, hash_: str):
        return self._entity_caps_cache.get((hash_method, hash_))

    @timeit
    def update_caps_time(self, method: str, hash_: str) -> None:
        sql = """UPDATE caps_cache SET last_seen = ?
                 WHERE hash_method = ? and hash = ?"""
        self._con.execute(sql, (int(time.time()), method, hash_))
        self._delayed_commit()

    @timeit
    def _clean_caps_table(self) -> None:
        """
        Remove caps which was not seen for 3 months
        """
        timestamp = int(time.time()) - 3 * 30 * 24 * 3600
        self._con.execute("DELETE FROM caps_cache WHERE last_seen < ?", (timestamp,))
        self._delayed_commit()

    @timeit
    def _fill_disco_info_cache(self) -> None:
        sql = """SELECT disco_info as "disco_info [disco_info]",
                 jid as "jid [jid]", last_seen FROM
                 last_seen_disco_info"""
        rows = self._con.execute(sql).fetchall()
        for row in rows:
            disco_info = row.disco_info._replace(timestamp=row.last_seen)
            self._disco_info_cache[row.jid] = disco_info
        log.info("%d DiscoInfo entries loaded", len(rows))

    def get_last_disco_info(self, jid: JID, max_age: int = 0) -> DiscoInfo | None:
        """
        Get last disco info from jid

        :param jid:         The jid

        :param max_age:     max age in seconds of the DiscoInfo record

        """

        disco_info = self._disco_info_cache.get(jid)
        if disco_info is not None:
            max_timestamp = time.time() - max_age if max_age else 0
            if max_timestamp > disco_info.timestamp:  # type: ignore
                return None
        return disco_info

    @timeit
    def set_last_disco_info(
        self, jid: JID, disco_info: DiscoInfo, cache_only: bool = False
    ) -> None:
        """
        Get last disco info from jid

        :param jid:          The jid

        :param disco_info:   A DiscoInfo object

        """

        log.info("Save disco info from %s", jid)

        if cache_only:
            self._disco_info_cache[jid] = disco_info
            return

        disco_exists = self.get_last_disco_info(jid) is not None
        if disco_exists:
            sql = """UPDATE last_seen_disco_info SET
                     disco_info = ?, last_seen = ?
                     WHERE jid = ?"""

            self._con.execute(sql, (disco_info, disco_info.timestamp, str(jid)))

        else:
            sql = """INSERT INTO last_seen_disco_info
                     (jid, disco_info, last_seen)
                     VALUES (?, ?, ?)"""

            self._con.execute(sql, (str(jid), disco_info, disco_info.timestamp))

        self._disco_info_cache[jid] = disco_info
        self._delayed_commit()

    @timeit
    def store_roster(self, account: str, roster: dict[JID, RosterItem]) -> None:
        serialized = json.dumps(list(roster.values()), cls=Encoder)

        insert_sql = "INSERT INTO roster(account, roster) VALUES(?, ?)"
        update_sql = "UPDATE roster SET roster = ? WHERE account = ?"

        try:
            self._con.execute(insert_sql, (account, serialized))
        except sqlite3.IntegrityError:
            self._con.execute(update_sql, (serialized, account))

        self._delayed_commit()

    @timeit
    def load_roster(self, account: str) -> dict[JID, RosterItem] | None:
        select_sql = "SELECT roster FROM roster WHERE account = ?"
        result = self._con.execute(select_sql, (account,)).fetchone()
        if result is None:
            return None

        roster: dict[JID, RosterItem] = {}
        data = json.loads(result.roster, object_hook=json_decoder)
        for item in data:
            roster[item.jid] = item
        return roster

    @timeit
    def remove_roster(self, account: str) -> None:
        delete_sql = "DELETE FROM roster WHERE account = ?"
        self._con.execute(delete_sql, (account,))
        self._commit()

    @timeit
    def get_unread(self) -> list[UnreadTableRow]:
        sql = "SELECT * FROM unread"
        return self._con.execute(sql).fetchall()

    @timeit
    def get_unread_count(self, account: str, jid: JID) -> int | None:
        sql = """SELECT count, message_id, timestamp FROM unread
                 WHERE account = ? AND jid = ?"""
        return self._con.execute(sql, (account, jid)).fetchone()

    @timeit
    def set_unread_count(
        self, account: str, jid: JID, count: int, message_id: str, timestamp: float
    ) -> None:
        if self.get_unread_count(account, jid) is not None:
            self.update_unread_count(account, jid, count)
        else:
            sql = """INSERT INTO unread
                     (account, jid, count, message_id, timestamp)
                     VALUES (?, ?, ?, ?, ?)"""
            self._con.execute(sql, (account, jid, count, message_id, timestamp))
            self._delayed_commit()

    @timeit
    def update_unread_count(self, account: str, jid: JID, count: int) -> None:
        sql = "UPDATE unread SET count = ? WHERE account = ? AND jid = ?"
        self._con.execute(sql, (count, account, jid))
        self._delayed_commit()

    @timeit
    def reset_unread_count(self, account: str, jid: JID) -> None:
        sql = "DELETE FROM unread WHERE account = ? AND jid = ?"
        self._con.execute(sql, (account, jid))
        self._delayed_commit()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import signal


def init_process() -> None:
    from gajim.main import set_proc_title

    set_proc_title("gajim-worker")

    signal.signal(signal.SIGINT, signal.SIG_IGN)

    gi_require_versions()


def gi_require_versions() -> None:
    import gi

    gi.require_versions(
        {
            "GLib": "2.0",
            "GdkPixbuf": "2.0",
        }
    )



# Copyright (C) 2005 Travis Shirk <travis AT pobox.com>
#                    Vincent Hanquez <tab AT snarc.org>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2006-2007 Jean-Marie Traissard <jim AT lapin.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


from enum import IntEnum
from enum import unique

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.setting_values import ADVANCED_SETTINGS
from gajim.common.setting_values import APP_SETTINGS

from gajim.gtk.builder import get_builder
from gajim.gtk.window import GajimAppWindow


@unique
class Column(IntEnum):
    NAME = 0
    VALUE = 1
    TYPE = 2
    IS_DEFAULT = 3


BOOL_DICT = {True: _("Activated"), False: _("Deactivated")}


SETTING_TYPES = {
    bool: p_("Configuration type", "Boolean"),
    int: p_("Configuration type", "Integer"),
    str: p_("Configuration type", "Text"),
}


class AdvancedConfig(GajimAppWindow):
    def __init__(self, transient_for: Gtk.Window | None) -> None:
        GajimAppWindow.__init__(
            self,
            name="AdvancedConfig",
            title=_("Advanced Configuration Editor (ACE)"),
            modal=True,
            transient_for=transient_for,
            add_window_padding=True,
            header_bar=True,
        )

        self._ui = get_builder("advanced_configuration.ui")

        self.set_child(self._ui.box)

        self._connect(self._ui.reset_button, "clicked", self._on_reset_button_clicked)
        self._connect(
            self._ui.search_entry, "search-changed", self._on_search_entry_changed
        )
        self._connect(
            self._ui.advanced_treeview, "row-activated", self._on_treeview_row_activated
        )
        self._connect(
            self._ui.treeview_selection, "changed", self._on_treeview_selection_changed
        )

        treeview = self._ui.advanced_treeview
        self.treeview = treeview
        self.model = Gtk.TreeStore(str, str, str, bool)

        self._fill_model()
        self.model.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.modelfilter = self.model.filter_new()
        self.modelfilter.set_visible_func(self._visible_func, self._ui.search_entry)

        renderer_text = Gtk.CellRendererText()
        renderer_text.set_property("ellipsize", Pango.EllipsizeMode.END)
        col = Gtk.TreeViewColumn(
            p_("Configuration", "Preference Name"), renderer_text, text=0
        )
        treeview.insert_column(col, -1)
        col.set_cell_data_func(renderer_text, self._value_column_name_callback)
        col.set_expand(True)
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        col.set_resizable(True)

        self.renderer_text = Gtk.CellRendererText()
        self.renderer_text.set_property("ellipsize", Pango.EllipsizeMode.END)
        self._connect(self.renderer_text, "edited", self._on_config_edited)

        col = Gtk.TreeViewColumn(
            p_("Configuration", "Value"), self.renderer_text, text=1
        )
        treeview.insert_column(col, -1)
        col.set_cell_data_func(self.renderer_text, self._value_column_data_callback)
        col.set_expand(True)
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        col.set_resizable(True)

        renderer_text = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn(p_("Configuration", "Type"), renderer_text, text=2)
        treeview.insert_column(col, -1)
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)

        treeview.set_model(self.modelfilter)

        self._connect(
            self.get_default_controller(), "key-pressed", self._on_key_pressed
        )

    def _cleanup(self) -> None:
        del self.renderer_text
        del self.modelfilter
        del self.model
        del self.treeview

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        _state: Gdk.ModifierType,
    ) -> bool:
        if keyval != Gdk.KEY_Escape:
            return Gdk.EVENT_PROPAGATE

        if self._ui.search_entry.get_text():
            self._ui.search_entry.set_text("")
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    @staticmethod
    def _value_column_name_callback(
        _col: Gtk.TreeViewColumn,
        cell: Gtk.CellRenderer,
        model: Gtk.TreeModel,
        iter_: Gtk.TreeIter,
        _data: object | None,
    ) -> None:
        opt_is_default = model[iter_][Column.IS_DEFAULT]
        cell.set_property("weight", 400 if opt_is_default else 700)

    @staticmethod
    def _value_column_data_callback(
        _col: Gtk.TreeViewColumn,
        cell: Gtk.CellRenderer,
        model: Gtk.TreeModel,
        iter_: Gtk.TreeIter,
        _data: object | None,
    ) -> None:
        opt_type = model[iter_][Column.TYPE]
        cell.set_property("editable", opt_type != SETTING_TYPES[bool])

        opt_is_default = model[iter_][Column.IS_DEFAULT]
        cell.set_property("weight", 400 if opt_is_default else 700)

    def _on_treeview_selection_changed(self, treeselection: Gtk.TreeSelection) -> None:
        model, iter_ = treeselection.get_selected()
        if not iter_:
            self._ui.reset_button.set_sensitive(False)
            return

        setting = model[iter_][Column.NAME]
        desc = ADVANCED_SETTINGS["app"][setting]

        self._ui.description.set_text(desc or p_("Configuration description", "None"))
        self._ui.reset_button.set_sensitive(not model[iter_][Column.IS_DEFAULT])

    def _on_treeview_row_activated(
        self, _treeview: Gtk.TreeView, path: Gtk.TreePath, _column: Gtk.TreeViewColumn
    ) -> None:
        modelpath = self.modelfilter.convert_path_to_child_path(path)
        assert modelpath
        modelrow = self.model[modelpath]
        setting = modelrow[Column.NAME]

        if modelrow[Column.TYPE] != SETTING_TYPES[bool]:
            return

        setting_value = modelrow[Column.VALUE] != _("Activated")
        column_value = BOOL_DICT[setting_value]
        default = APP_SETTINGS[setting]

        app.settings.set(setting, setting_value)
        modelrow[Column.VALUE] = column_value
        modelrow[Column.IS_DEFAULT] = bool(setting_value == default)

        self._ui.reset_button.set_sensitive(setting_value != default)

    def _on_config_edited(
        self, _cell: Gtk.CellRendererText, path: str, text: str
    ) -> None:
        treepath = Gtk.TreePath.new_from_string(path)
        assert treepath is not None
        modelpath = self.modelfilter.convert_path_to_child_path(treepath)
        assert modelpath
        modelrow = self.model[modelpath]
        setting = modelrow[Column.NAME]
        default = APP_SETTINGS[setting]

        value = text
        if modelrow[Column.TYPE] == SETTING_TYPES[int]:
            try:
                value = int(text)
            except ValueError:
                return

        app.settings.set(setting, value)
        modelrow[Column.VALUE] = text
        modelrow[Column.IS_DEFAULT] = bool(value == default)

        self._ui.reset_button.set_sensitive(value != default)

    def _on_reset_button_clicked(self, button: Gtk.Button) -> None:
        model, iter_ = self.treeview.get_selection().get_selected()
        if not iter_:
            return

        setting = model[iter_][Column.NAME]
        default = APP_SETTINGS[setting]

        if isinstance(default, bool):
            model[iter_][Column.VALUE] = BOOL_DICT[default]
        else:
            model[iter_][Column.VALUE] = str(default)

        model[iter_][Column.IS_DEFAULT] = True

        app.settings.set(setting, default)
        button.set_sensitive(False)

    def _fill_model(self) -> None:
        for category, settings in ADVANCED_SETTINGS.items():
            if category != "app":
                continue

            for setting in settings:
                value = app.settings.get(setting)  # type: ignore
                default = APP_SETTINGS[setting]
                is_default = bool(value == default)  # type: ignore

                if isinstance(value, bool):
                    value = BOOL_DICT[value]
                    type_ = SETTING_TYPES[bool]

                elif isinstance(value, int):
                    value = str(value)
                    type_ = SETTING_TYPES[int]

                elif isinstance(value, str):
                    type_ = SETTING_TYPES[str]

                else:
                    raise ValueError

                self.model.append(None, [setting, value, type_, is_default])

    @staticmethod
    def _visible_func(
        model: Gtk.TreeModel, treeiter: Gtk.TreeIter, search_entry: Gtk.SearchEntry
    ) -> bool:
        search_string = search_entry.get_text().lower()
        if not search_string:
            return True

        setting = model[treeiter][Column.NAME]
        desc = ADVANCED_SETTINGS["app"][setting]

        return search_string in setting or search_string in desc.lower()

    def _on_search_entry_changed(self, _entry: Gtk.SearchEntry) -> None:
        self.modelfilter.refilter()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging

from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.gstreamer import create_video_elements

try:
    from gi.repository import Gst
except Exception:
    if typing.TYPE_CHECKING:
        from gi.repository import Gst


log = logging.getLogger("gajim.gtk.preview")


class VideoPreview(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(self)

        self._active = False

        self._av_pipeline: Gst.Pipeline | None = None
        self._av_src: Gst.Bin | None = None
        self._av_sink: Gst.Element | None = None
        self._av_widget: Gtk.Widget | None = None

        self._ui = get_builder("video_preview.ui")
        self.append(self._ui.video_preview_box)

    def do_unroot(self) -> None:
        self._disable_preview()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    @property
    def is_active(self) -> bool:
        return self._active

    def toggle_preview(self, value: bool) -> None:
        self._active = value
        if value:
            return self._enable_preview()
        return self._disable_preview()

    def _enable_preview(self) -> None:
        src_name = app.settings.get("video_input_device")
        try:
            self._av_src = Gst.parse_bin_from_description(src_name, True)
        except GLib.Error as error:
            log.error(error)
            log.error('Failed to parse "%s" as Gstreamer element', src_name)
            self._set_error_text()
            return

        video_elements = create_video_elements()
        if video_elements is None:
            log.error(
                "Failed to obtain a working Gstreamer GTK+ sink, "
                "video support will be disabled"
            )
            self._set_error_text()
            return

        sink, paintable, name = video_elements
        self._set_sink_text(name)

        if self._av_pipeline is None:
            self._av_pipeline = Gst.Pipeline.new("preferences-pipeline")
        else:
            self._av_pipeline.set_state(Gst.State.NULL)

        self._av_pipeline.add(sink)
        self._av_sink = sink

        if self._av_widget is not None:
            self._ui.video_preview_box.remove(self._av_widget)

        self._ui.video_preview_placeholder.set_visible(False)
        self._av_widget = Gtk.Picture(paintable=paintable)
        self._ui.video_preview_box.append(self._av_widget)

        assert self._av_src is not None
        self._av_pipeline.add(self._av_src)
        self._av_src.link(self._av_sink)
        self._av_pipeline.set_state(Gst.State.PLAYING)

    def _disable_preview(self) -> None:
        if self._av_pipeline is not None:
            self._av_pipeline.set_state(Gst.State.NULL)
            if self._av_src is not None:
                self._av_pipeline.remove(self._av_src)
            if self._av_sink is not None:
                self._av_pipeline.remove(self._av_sink)

        self._av_src = None
        self._av_sink = None

        if self._av_widget is not None:
            self._ui.video_preview_box.remove(self._av_widget)
            self._ui.video_preview_placeholder.set_visible(True)
            self._av_widget = None
        self._av_pipeline = None

        self._ui.video_source_label.set_visible(False)

    def _set_sink_text(self, sink_name: str) -> None:
        label_markup = '<span color="%s" font-weight="bold">%s</span>'
        color = "black"
        label_text = ""
        if sink_name == "gtkglsink":
            color = "green"
            label_text = _("OpenGL accelerated")

        elif sink_name == "gtksink":
            color = "orange"
            label_text = _("Not accelerated")

        label_markup = label_markup % (color, label_text)
        self._ui.video_source_label.set_markup(label_markup)
        self._ui.video_source_label.set_visible(True)

    def _set_error_text(self) -> None:
        self._ui.video_source_label.set_text(
            _("Something went wrong. Video feature disabled.")
        )
        self._ui.video_source_label.set_visible(True)

    def refresh(self) -> None:
        self.toggle_preview(False)
        self.toggle_preview(True)



# Copyright (C) 2006 Dimitur Kirov <dkirov AT gmail.com>
#                    Junglecow J <junglecow AT gmail.com>
# Copyright (C) 2006-2007 Tomasz Melcer <liori AT exroot.org>
#                         Travis Shirk <travis AT pobox.com>
#                         Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
#                         Jean-Marie Traissard <jim AT lapin.org>
#                         Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging
import socket

import nbxmpp
from gi.repository import GLib
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import helpers
from gajim.common import types
from gajim.common.events import FileRequestError
from gajim.common.file_props import FileProp
from gajim.common.file_props import FilesProp
from gajim.common.modules.base import BaseModule

# from gajim.common.socks5 import Socks5SenderClient
from gajim.common.util.jid import get_full_jid_from_iq
from gajim.common.util.jid import get_jid_from_iq
from gajim.common.util.jid import parse_jid

log = logging.getLogger("gajim.c.m.bytestream")


def is_transfer_paused(file_props: FileProp) -> bool:
    if file_props.stopped:
        return False
    if file_props.completed:
        return False
    if file_props.disconnect_cb:
        return False
    return file_props.paused


def is_transfer_active(file_props: FileProp) -> bool:
    if file_props.stopped:
        return False
    if file_props.completed:
        return False
    if not file_props.started:
        return False
    if file_props.paused:
        return True
    return not file_props.paused


def is_transfer_stopped(file_props: FileProp) -> bool:
    if not file_props:
        return True
    if file_props.error:
        return True
    if file_props.completed:
        return True
    if not file_props.stopped:  # noqa: SIM103
        return False
    return True


class Bytestream(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="iq",
                typ="result",
                ns=Namespace.BYTESTREAM,
                callback=self._on_bytestream_result,
            ),
            StanzaHandler(
                name="iq",
                typ="error",
                ns=Namespace.BYTESTREAM,
                callback=self._on_bytestream_error,
            ),
            StanzaHandler(
                name="iq",
                typ="set",
                ns=Namespace.BYTESTREAM,
                callback=self._on_bytestream_set,
            ),
            StanzaHandler(name="iq", typ="result", callback=self._on_result),
        ]

        self.no_gupnp_reply_id = None
        self.ok_id = None
        self.fail_id = None

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.BYTESTREAM not in info.features:
            return
        if app.settings.get_account_setting(self._account, "use_ft_proxies"):
            log.info("Discovered proxy: %s", info.jid)
            our_fjid = self._con.get_own_jid()
            testit = app.settings.get_account_setting(
                self._account, "test_ft_proxies_on_startup"
            )
            raise nbxmpp.NodeProcessed

    @staticmethod
    def _ft_get_receiver_jid(file_props: FileProp) -> str:
        return file_props.receiver.jid + "/" + file_props.receiver.resource

    @staticmethod
    def _ft_get_from(iq_obj: Iq) -> str | None:
        return get_full_jid_from_iq(iq_obj)

    @staticmethod
    def _ft_get_streamhost_jid_attr(streamhost: nbxmpp.Node) -> str:
        return parse_jid(streamhost.getAttr("jid"))

    def send_file_approval(self, file_props: FileProp) -> None:
        """
        Send iq, confirming that we want to download the file
        """
        # user response to ConfirmationAlertDialog may come after we've disconnected
        if not app.account_is_available(self._account):
            return

        # file transfer initiated by a jingle session
        log.info("send_file_approval: jingle session accept")

        session = self._con.get_module("Jingle").get_jingle_session(
            file_props.sender, file_props.sid
        )
        if not session:
            return
        content = None
        for content_ in session.contents.values():
            if content_.transport.sid == file_props.transport_sid:
                content = content_
                break

        if not content:
            return

        if not session.accepted:
            content = session.get_content("file", content.name)
            session.approve_session()

        session.approve_content("file", content.name)

    def send_file_rejection(self, file_props: FileProp) -> None:
        """
        Inform sender that we refuse to download the file

        typ is used when code = '400', in this case typ can be 'stream' for
        invalid stream or 'profile' for invalid profile
        """
        # user response to ConfirmationAlertDialog may come after we've disconnected
        if not app.account_is_available(self._account):
            return

        for session in self._con.get_module("Jingle").get_jingle_sessions(
            None, file_props.sid
        ):
            session.cancel_session()

    def send_success_connect_reply(self, streamhost: dict[str, Any] | None) -> None:
        """
        Send reply to the initiator of FT that we made a connection
        """
        if not app.account_is_available(self._account):
            return
        if streamhost is None:
            return
        iq = nbxmpp.Iq(
            to=streamhost["initiator"], typ="result", frm=streamhost["target"]
        )
        iq.setAttr("id", streamhost["id"])
        query = iq.setTag("query", namespace=Namespace.BYTESTREAM)
        stream_tag = query.setTag("streamhost-used")
        stream_tag.setAttr("jid", streamhost["jid"])
        self._con.connection.send(iq)

    def stop_all_active_file_transfers(self, contact: types.ChatContactT):
        """
        Stop all active transfer to or from the given contact
        """
        for file_props in FilesProp.getAllFileProp():
            if is_transfer_stopped(file_props):
                continue
            receiver_jid = file_props.receiver
            if contact.jid == receiver_jid:
                file_props.error = -5
                self.remove_transfer(file_props)
                app.ged.raise_event(
                    FileRequestError(
                        conn=self._con,
                        jid=contact.jid.bare,
                        file_props=file_props,
                        error_msg="",
                    )
                )
            sender_jid = file_props.sender
            if contact.jid == sender_jid:
                file_props.error = -3
                self.remove_transfer(file_props)

    def remove_all_transfers(self) -> None:
        """
        Stop and remove all active connections from the socks5 pool
        """
        for file_props in FilesProp.getAllFileProp():
            self.remove_transfer(file_props)

    def remove_transfer(self, file_props: FileProp | None) -> None:
        if file_props is None:
            return
        self.disconnect_transfer(file_props)

    @staticmethod
    def disconnect_transfer(file_props: FileProp) -> None:
        if file_props is None:
            return
        if file_props.hash_:
            app.socks5queue.remove_sender(file_props.hash_)

        if file_props.streamhosts:
            for host in file_props.streamhosts:
                if "idx" in host and host["idx"] > 0:
                    app.socks5queue.remove_receiver(host["idx"])
                    app.socks5queue.remove_sender(host["idx"])

    def _send_socks5_info(self, file_props: FileProp) -> None:
        """
        Send iq for the present streamhosts and proxies
        """
        if not app.account_is_available(self._account):
            return
        receiver = file_props.receiver
        sender = file_props.sender

        sha_str = helpers.get_auth_sha(file_props.sid, sender, receiver)
        file_props.sha_str = sha_str

        port = app.settings.get("file_transfers_port")
        listener = app.socks5queue.start_listener(
            port, sha_str, self._result_socks5_sid, file_props
        )
        if not listener:
            file_props.error = -5
            app.ged.raise_event(
                FileRequestError(
                    conn=self._con,
                    jid=app.get_jid_without_resource(receiver),
                    file_props=file_props,
                    error_msg="",
                )
            )
            self._connect_error(
                file_props.sid, error="not-acceptable", error_type="modify"
            )
        else:
            iq = nbxmpp.Iq(to=receiver, typ="set")
            file_props.request_id = "id_" + file_props.sid
            iq.setID(file_props.request_id)
            query = iq.setTag("query", namespace=Namespace.BYTESTREAM)
            query.setAttr("sid", file_props.sid)

            self._add_addiditional_streamhosts_to_query(query, file_props)
            self._add_local_ips_as_streamhosts_to_query(query, file_props)
            self._add_proxy_streamhosts_to_query(query, file_props)
            self._add_upnp_igd_as_streamhost_to_query(query, file_props, iq)
            # Upnp-igd is asynchronous, so it will send the iq itself

    @staticmethod
    def _add_streamhosts_to_query(
        query: nbxmpp.Node, sender: str, port: str, hosts: list[str]
    ) -> None:
        for host in hosts:
            streamhost = nbxmpp.Node(tag="streamhost")
            query.addChild(node=streamhost)
            streamhost.setAttr("port", str(port))
            streamhost.setAttr("host", host)
            streamhost.setAttr("jid", sender)

    def _add_local_ips_as_streamhosts_to_query(
        self, query: nbxmpp.Node, file_props: FileProp
    ) -> None:
        if not app.settings.get_account_setting(self._account, "ft_send_local_ips"):
            return

        my_ip = self._con.local_address
        if my_ip is None:
            log.warning("No local address available")
            return

        try:
            # The ip we're connected to server with
            my_ips = [my_ip]
            # all IPs from local DNS
            for addr in socket.getaddrinfo(socket.gethostname(), None):
                if (
                    addr[4][0] not in my_ips
                    and not addr[4][0].startswith("127")
                    and not addr[4][0] == "::1"
                ):
                    my_ips.append(addr[4][0])

            sender = file_props.sender
            port = app.settings.get("file_transfers_port")
            self._add_streamhosts_to_query(query, sender, port, my_ips)
        except socket.gaierror:
            log.error("wrong host, invalid local address?")

    def _add_addiditional_streamhosts_to_query(
        self, query: nbxmpp.Node, file_props: FileProp
    ) -> None:
        sender = file_props.sender
        port = app.settings.get("file_transfers_port")
        ft_add_hosts_to_send = app.settings.get("ft_add_hosts_to_send")
        add_hosts = []
        if ft_add_hosts_to_send:
            add_hosts = [e.strip() for e in ft_add_hosts_to_send.split(",")]
        else:
            add_hosts = []
        self._add_streamhosts_to_query(query, sender, port, add_hosts)

    def _add_upnp_igd_as_streamhost_to_query(
        self, query: nbxmpp.Node, file_props: FileProp, iq: Iq
    ) -> None:
        my_ip = self._con.local_address
        if my_ip is None or not app.is_installed("UPNP"):
            log.warning("No local address available")
            self._con.connection.send(iq)
            return

        # check if we are connected with an IPv4 address
        try:
            socket.inet_aton(my_ip)
        except OSError:
            self._con.connection.send(iq)
            return

        def ip_is_local(ip: str) -> bool:
            if "." not in ip:
                # it's an IPv6
                return True
            ip_s = ip.split(".")
            ip_l = (
                int(ip_s[0]) << 24
                | int(ip_s[1]) << 16
                | int(ip_s[2]) << 8
                | int(ip_s[3])
            )
            # 10/8
            if ip_l & (255 << 24) == 10 << 24:
                return True
            # 172.16/12
            if ip_l & (255 << 24 | 240 << 16) == (172 << 24 | 16 << 16):
                return True
            # 192.168
            if ip_l & (255 << 24 | 255 << 16) == (192 << 24 | 168 << 16):  # noqa: SIM103
                return True
            return False

        if not ip_is_local(my_ip):
            self.connection.send(iq)
            return

        self.no_gupnp_reply_id = 0

        def cleanup_gupnp() -> None:
            if self.no_gupnp_reply_id:
                GLib.source_remove(self.no_gupnp_reply_id)
                self.no_gupnp_reply_id = 0
            app.gupnp_igd.disconnect(self.ok_id)
            app.gupnp_igd.disconnect(self.fail_id)

        def success(_gupnp, _proto, ext_ip, _re, ext_port, local_ip, local_port, _desc):
            log.debug(
                "Got GUPnP-IGD answer: external: %s:%s, internal: %s:%s",
                ext_ip,
                ext_port,
                local_ip,
                local_port,
            )
            if local_port != app.settings.get("file_transfers_port"):
                sender = file_props.sender
                receiver = file_props.receiver
                sha_str = helpers.get_auth_sha(file_props.sid, sender, receiver)
                listener = app.socks5queue.start_listener(
                    local_port, sha_str, self._result_socks5_sid, file_props.sid
                )
                if listener:
                    self._add_streamhosts_to_query(query, sender, ext_port, [ext_ip])
            else:
                self._add_streamhosts_to_query(
                    query, file_props.sender, ext_port, [ext_ip]
                )
            self._con.connection.send(iq)
            cleanup_gupnp()

        def fail(_gupnp, error, _proto, _ext_ip, _local_ip, _local_port, _desc):
            log.debug("Got GUPnP-IGD error: %s", error)
            self._con.connection.send(iq)
            cleanup_gupnp()

        def no_upnp_reply() -> bool:
            log.debug("Got not GUPnP-IGD answer")
            # stop trying to use it
            app.disable_dependency("UPNP")
            self.no_gupnp_reply_id = 0
            self._con.connection.send(iq)
            cleanup_gupnp()
            return False

        self.ok_id = app.gupnp_igd.connect("mapped-external-port", success)
        self.fail_id = app.gupnp_igd.connect("error-mapping-port", fail)

        port = app.settings.get("file_transfers_port")
        self.no_gupnp_reply_id = GLib.timeout_add_seconds(10, no_upnp_reply)
        app.gupnp_igd.add_port("TCP", 0, my_ip, port, 3600, "Gajim file transfer")

    def _add_proxy_streamhosts_to_query(
        self, query: nbxmpp.Node, file_props: FileProp
    ) -> None:
        proxyhosts = self._get_file_transfer_proxies_from_config(file_props)
        if proxyhosts:
            file_props.proxy_receiver = file_props.receiver
            file_props.proxy_sender = file_props.sender
            file_props.proxyhosts = proxyhosts

            for proxyhost in proxyhosts:
                self._add_streamhosts_to_query(
                    query, proxyhost["jid"], proxyhost["port"], [proxyhost["host"]]
                )

    def _get_file_transfer_proxies_from_config(
        self, file_props: FileProp
    ) -> list[dict[str, Any]]:
        return []

        # configured_proxies = app.settings.get_account_setting(
        #     self._account, 'file_transfer_proxies')
        # shall_use_proxies = app.settings.get_account_setting(
        #     self._account, 'use_ft_proxies')
        # if shall_use_proxies:
        #     proxyhost_dicts: list[dict[str, Any]] = []
        #     proxies: list[str] = []
        #     if configured_proxies:
        #         proxies = [item.strip() for item in
        #                    configured_proxies.split(',')]
        #     default_proxy = app.proxy65_manager.get_default_for_name(
        #         self._account)
        #     if default_proxy:
        #         # add/move default proxy at top of the others
        #         if default_proxy in proxies:
        #             proxies.remove(default_proxy)
        #         proxies.insert(0, default_proxy)

        #     for proxy in proxies:
        #         (host, _port, jid) = app.proxy65_manager.get_proxy(
        #             proxy, self._account)
        #         if not host:
        #             continue
        #         host_dict: dict[str, Any] = {
        #             'state': 0,
        #             'target': file_props.receiver,
        #             'id': file_props.sid,
        #             'sid': file_props.sid,
        #             'initiator': proxy,
        #             'host': host,
        #             'port': str(_port),
        #             'jid': jid
        #         }
        #         proxyhost_dicts.append(host_dict)
        #     return proxyhost_dicts

        # return []

    @staticmethod
    def _result_socks5_sid(sid: str, hash_id: str) -> None:
        """
        Store the result of SHA message from auth
        """
        file_props = FilesProp.getFilePropBySid(sid)
        file_props.hash_ = hash_id

    def _connect_error(
        self, sid: str, error: str, error_type: str, msg: str | None = None
    ) -> None:
        """
        Called when there is an error establishing BS connection, or when
        connection is rejected
        """
        if not app.account_is_available(self._account):
            return
        file_props = FilesProp.getFileProp(self._account, sid)
        if file_props is None:
            log.error("can not send iq error on failed transfer")
            return
        if file_props.type_ == "s":
            to = file_props.receiver
        else:
            to = file_props.sender
        iq = nbxmpp.Iq(to=to, typ="error")
        iq.setAttr("id", file_props.request_id)
        err = iq.setTag("error")
        err.setAttr("type", error_type)
        err.setTag(error, namespace=Namespace.STANZAS)
        self._con.connection.send(iq)
        if msg:
            self.disconnect_transfer(file_props)
            file_props.error = -3
            app.ged.raise_event(
                FileRequestError(
                    conn=self._con,
                    jid=app.get_jid_without_resource(to),
                    file_props=file_props,
                    error_msg=msg,
                )
            )

    def _proxy_auth_ok(self, proxy: dict[str, Any]) -> None:
        """
        Called after authentication to proxy server
        """
        if not app.account_is_available(self._account):
            return
        file_props = FilesProp.getFileProp(self._account, proxy["sid"])
        iq = nbxmpp.Iq(to=proxy["initiator"], typ="set")
        auth_id = "au_" + proxy["sid"]
        iq.setID(auth_id)
        query = iq.setTag("query", namespace=Namespace.BYTESTREAM)
        query.setAttr("sid", proxy["sid"])
        activate = query.setTag("activate")
        activate.setData(file_props.proxy_receiver)
        iq.setID(auth_id)
        self._con.connection.send(iq)

    def _on_bytestream_error(
        self, _con: types.NBXMPPClient, iq_obj: Iq, _properties: IqProperties
    ) -> None:
        id_ = iq_obj.getAttr("id")
        frm = get_full_jid_from_iq(iq_obj)
        query = iq_obj.getTag("query")
        jid = get_jid_from_iq(iq_obj)
        id_ = id_[3:]
        file_props = FilesProp.getFilePropBySid(id_)
        if not file_props:
            return
        file_props.error = -4
        app.ged.raise_event(
            FileRequestError(
                conn=self._con,
                jid=app.get_jid_without_resource(jid),
                file_props=file_props,
                error_msg="",
            )
        )
        raise nbxmpp.NodeProcessed

    def _on_bytestream_set(
        self, _con: types.NBXMPPClient, iq_obj: Iq, _properties: IqProperties
    ):
        target = iq_obj.getAttr("to")
        id_ = iq_obj.getAttr("id")
        query = iq_obj.getTag("query")
        sid = query.getAttr("sid")
        streamhosts: list[dict[str, Any]] = []
        for item in query.getChildren():
            if item.getName() == "streamhost":
                host_dict = {
                    "state": 0,
                    "target": target,
                    "id": id_,
                    "sid": sid,
                    "initiator": self._ft_get_from(iq_obj),
                }
                for attr in item.getAttrs():
                    host_dict[attr] = item.getAttr(attr)
                if "host" not in host_dict:
                    continue
                if "jid" not in host_dict:
                    continue
                if "port" not in host_dict:
                    continue
                streamhosts.append(host_dict)
        file_props = FilesProp.getFilePropBySid(sid)
        if file_props is not None:
            if file_props.type_ == "s":  # FIXME: remove fast xmlns
                # only psi do this
                if file_props.streamhosts:
                    file_props.streamhosts.extend(streamhosts)
                else:
                    file_props.streamhosts = streamhosts
                app.socks5queue.connect_to_hosts(
                    self._account, sid, self.send_success_connect_reply, None
                )
                raise nbxmpp.NodeProcessed
        else:
            log.warning("Gajim got streamhosts for unknown transfer. Ignoring it.")
            raise nbxmpp.NodeProcessed

        file_props.streamhosts = streamhosts

        def _connection_error(sid: str) -> None:
            self._connect_error(
                sid, "item-not-found", "cancel", msg="Could not connect to given hosts"
            )

        if file_props.type_ == "r":
            app.socks5queue.connect_to_hosts(
                self._account, sid, self.send_success_connect_reply, _connection_error
            )
        raise nbxmpp.NodeProcessed

    def _on_result(
        self, _con: types.NBXMPPClient, iq_obj: Iq, _properties: IqProperties
    ) -> None:
        # if we want to respect xep-0065 we have to check for proxy
        # activation result in any result iq
        real_id = iq_obj.getAttr("id")
        if real_id is None:
            log.warning("Invalid IQ without id attribute:\n%s", iq_obj)
            raise nbxmpp.NodeProcessed
        if real_id is None or not real_id.startswith("au_"):
            return
        frm = self._ft_get_from(iq_obj)
        id_ = real_id[3:]
        file_props = FilesProp.getFilePropByTransportSid(self._account, id_)
        if file_props.streamhost_used:
            for host in file_props.proxyhosts:
                if host["initiator"] == frm and "idx" in host:
                    app.socks5queue.activate_proxy(host["idx"])
                    raise nbxmpp.NodeProcessed

    def _on_bytestream_result(
        self, _con: types.NBXMPPClient, iq_obj: Iq, _properties: IqProperties
    ) -> None:
        frm = self._ft_get_from(iq_obj)
        real_id = iq_obj.getAttr("id")
        query = iq_obj.getTag("query")

        try:
            streamhost = query.getTag("streamhost-used")
        except Exception:  # this bytestream result is not what we need
            pass
        id_ = real_id[3:]
        file_props = FilesProp.getFileProp(self._account, id_)
        if file_props is None:
            raise nbxmpp.NodeProcessed
        if streamhost is None:
            # proxy approves the activate query
            if real_id.startswith("au_"):
                if file_props.streamhost_used is False:
                    raise nbxmpp.NodeProcessed
                if not file_props.proxyhosts:
                    raise nbxmpp.NodeProcessed
                for host in file_props.proxyhosts:
                    if (
                        host["initiator"] == frm
                        and query.getAttr("sid") == file_props.sid
                    ):
                        app.socks5queue.activate_proxy(host["idx"])
                        break
            raise nbxmpp.NodeProcessed
        jid = self._ft_get_streamhost_jid_attr(streamhost)
        if file_props.streamhost_used is True:
            raise nbxmpp.NodeProcessed

        if real_id.startswith("au_"):
            if file_props.stopped:
                self.remove_transfer(file_props)
            else:
                app.socks5queue.send_file(file_props, self._account, "server")
            raise nbxmpp.NodeProcessed

        proxy = None
        if file_props.proxyhosts:
            for proxyhost in file_props.proxyhosts:
                if proxyhost["jid"] == jid:
                    proxy = proxyhost

        if file_props.stopped:
            self.remove_transfer(file_props)
            raise nbxmpp.NodeProcessed
        if proxy is not None:
            file_props.streamhost_used = True
            file_props.streamhosts.append(proxy)
            file_props.is_a_proxy = True
            idx = app.socks5queue.idx
            sender = Socks5SenderClient(
                app.idlequeue,
                idx,
                app.socks5queue,
                _sock=None,
                host=str(proxy["host"]),
                port=int(proxy["port"]),
                connected=False,
                file_props=file_props,
            )
            sender.streamhost = proxy
            app.socks5queue.add_sockobj(self._account, sender)
            proxy["idx"] = sender.queue_idx
            app.socks5queue.on_success[file_props.sid] = self._proxy_auth_ok
            raise nbxmpp.NodeProcessed

        if file_props.stopped:
            self.remove_transfer(file_props)
        else:
            app.socks5queue.send_file(file_props, self._account, "server")

        raise nbxmpp.NodeProcessed



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Constants for the gtk module

from __future__ import annotations

from typing import Any
from typing import NamedTuple

import sys
from collections.abc import Callable
from collections.abc import Iterator
from enum import Enum
from enum import IntEnum
from enum import unique

from gi.repository import Gdk

from gajim.common.i18n import _
from gajim.common.setting_values import AllSettingsT

if sys.byteorder == "little":
    gdk_memory_default = Gdk.MemoryFormat.B8G8R8A8_PREMULTIPLIED
else:
    gdk_memory_default = Gdk.MemoryFormat.A8R8G8B8_PREMULTIPLIED
GDK_MEMORY_DEFAULT = gdk_memory_default


class Setting(NamedTuple):
    kind: SettingKind
    label: str
    type: SettingType
    value: AllSettingsT | None = None
    name: str | None = None
    callback: Callable[..., None] | None = None
    data: Any | None = None
    desc: str | None = None
    bind: str | None = None
    inverted: bool | None = None
    enabled_func: Callable[..., bool] | None = None
    props: dict[str, Any] | None = None


DEFAULT_WORKSPACE_COLOR = "rgb(191,15,167)"

MAX_MESSAGE_LENGTH = 1000


@unique
class MuteState(IntEnum):
    UNMUTE = 0
    MIN_30 = 30
    MIN_60 = 60
    MIN_120 = 120
    MIN_480 = 480
    PERM = 100000000

    @classmethod
    def iter(cls) -> Iterator[tuple[int, str]]:
        yield from cls._labels.items()  # type: ignore


MuteState._labels = {  # type: ignore
    MuteState.MIN_30: _("30 minutes"),
    MuteState.MIN_60: _("1 hour"),
    MuteState.MIN_120: _("2 hours"),
    MuteState.MIN_480: _("8 hours"),
    MuteState.PERM: _("Permanently"),
}


@unique
class Theme(IntEnum):
    NOT_DARK = 0
    DARK = 1
    SYSTEM = 2


@unique
class SettingKind(IntEnum):
    ENTRY = 0
    SWITCH = 1
    SPIN = 2
    ACTION = 3
    DIALOG = 5
    CALLBACK = 6
    FILECHOOSER = 10
    COMBO = 12
    COLOR = 13
    DROPDOWN = 19
    GENERIC = 20
    SUBPAGE = 21


@unique
class SettingType(IntEnum):
    CONFIG = 0
    ACCOUNT_CONFIG = 1
    CONTACT = 2
    GROUP_CHAT = 3
    VALUE = 4
    ACTION = 5
    DIALOG = 6


class ControlType(Enum):
    CHAT = "chat"
    GROUPCHAT = "gc"
    PRIVATECHAT = "pm"

    @property
    def is_chat(self):
        return self == ControlType.CHAT

    @property
    def is_groupchat(self):
        return self == ControlType.GROUPCHAT

    @property
    def is_privatechat(self):
        return self == ControlType.PRIVATECHAT

    def __str__(self):
        return self.value


WINDOW_MODULES = {
    "AccountWizard": "gajim.gtk.account_wizard",
    "AddContact": "gajim.gtk.add_contact",
    "AdHocCommands": "gajim.gtk.adhoc",
    "AdvancedConfig": "gajim.gtk.advanced_config",
    "DBMigration": "gajim.gtk.db_migration",
    "DebugConsoleWindow": "gajim.gtk.debug_console",
    "CertificateDialog": "gajim.gtk.certificate_dialog",
    "ChangePassword": "gajim.gtk.change_password",
    "ContactInfo": "gajim.gtk.contact_info",
    "CreateGroupchatWindow": "gajim.gtk.groupchat_creation",
    "Features": "gajim.gtk.features",
    "GroupchatDetails": "gajim.gtk.groupchat_details",
    "GroupchatJoin": "gajim.gtk.groupchat_join",
    "HistoryExport": "gajim.gtk.history_export",
    "HistorySyncAssistant": "gajim.gtk.history_sync",
    "ManageProxies": "gajim.gtk.manage_proxies",
    "ManageSounds": "gajim.gtk.manage_sounds",
    "OpenPGPSecret": "gajim.gtk.openpgp_secret",
    "OpenPGPWizard": "gajim.gtk.openpgp_wizard",
    "PasswordDialog": "gajim.gtk.password_dialog",
    "PEPConfig": "gajim.gtk.pep_config",
    "Preferences": "gajim.gtk.preference.dialog",
    "ProfileWindow": "gajim.gtk.profile",
    "QuitDialog": "gajim.gtk.quit",
    "RemoveAccount": "gajim.gtk.remove_account",
    "RosterItemExchange": "gajim.gtk.roster_item_exchange",
    "ServiceDiscoveryWindow": "gajim.gtk.discovery",
    "ServiceRegistration": "gajim.gtk.service_registration",
    "SSLErrorDialog": "gajim.gtk.ssl_error_dialog",
    "StartChatDialog": "gajim.gtk.start_chat",
    "Themes": "gajim.gtk.themes",
    "WorkspaceDialog": "gajim.gtk.workspace_dialog",
}


APP_ACTIONS = [
    ("about", None),
    ("accounts", "s"),
    ("add-account", None),
    ("add-contact", "s"),
    ("content", None),
    ("copy-text", "s"),
    ("create-groupchat", "s"),
    ("faq", None),
    ("privacy-policy", None),
    ("features", None),
    ("forget-groupchat", "a{sv}"),
    ("join-support-chat", None),
    ("manage-proxies", None),
    ("open-link", "s"),
    ("plugins", None),
    ("preferences", None),
    ("quit", None),
    ("export-history", "a{sv}"),
    ("remove-history", "a{sv}"),
    ("shortcuts", None),
    ("show", None),
    ("start-chat", "as"),
    ("open-chat", "as"),
    ("mute-chat", "a{sv}"),
    ("xml-console", None),
    ("handle-uri", "as"),
    ("save-file-as", "s"),
    ("open-file", "s"),
    ("open-folder", "s"),
]


MAIN_WIN_ACTIONS = [
    # action name, variant type, enabled
    ("input-clear", None, True),
    ("input-bold", None, True),
    ("input-italic", None, True),
    ("input-strike", None, True),
    ("input-preformatted", None, True),
    ("input-preformatted-block", None, True),
    ("input-focus", None, True),
    ("input-emoji", None, True),
    ("input-paste-as-code-block", None, True),
    ("input-paste-as-quote", None, True),
    ("input-mention", "s", True),
    ("activate-message-selection", "u", True),
    ("delete-message-locally", "a{sv}", True),
    ("correct-message", None, False),
    ("retract-message", "a{sv}", False),
    ("copy-message", "s", True),
    ("moderate-message", "a{sv}", False),
    ("moderate-all-messages", "a{sv}", False),
    ("quote", "s", True),
    ("quote-next", None, True),
    ("quote-prev", None, True),
    ("reply", "u", True),
    ("jump-to-message", "au", True),
    ("send-file-httpupload", "as", False),
    # ('send-file-jingle', 'as', False),
    ("send-file", "as", False),
    ("start-voice-call", None, False),
    ("start-video-call", None, False),
    ("show-contact-info", None, True),
    ("show-chat-share", None, True),
    ("chat-contact-info", "a{sv}", True),
    ("send-message", None, False),
    ("muc-change-nickname", None, False),
    ("muc-invite", None, False),
    ("muc-contact-info", "s", False),
    ("muc-execute-command", "s", False),
    ("muc-ban", "s", False),
    ("muc-kick", "s", False),
    ("muc-change-role", "as", False),
    ("muc-change-affiliation", "as", False),
    ("muc-request-voice", None, False),
    ("muc-user-block", "a{sv}", True),
    ("muc-user-unblock", "a{sv}", True),
    ("scroll-view-up", None, True),
    ("scroll-view-down", None, True),
    ("change-nickname", None, True),
    ("change-subject", None, True),
    ("escape", None, True),
    ("close-chat", None, True),
    ("restore-chat", None, True),
    ("switch-next-chat", None, True),
    ("switch-prev-chat", None, True),
    ("switch-next-unread-chat", None, True),
    ("switch-prev-unread-chat", None, True),
    ("switch-chat-1", None, True),
    ("switch-chat-2", None, True),
    ("switch-chat-3", None, True),
    ("switch-chat-4", None, True),
    ("switch-chat-5", None, True),
    ("switch-chat-6", None, True),
    ("switch-chat-7", None, True),
    ("switch-chat-8", None, True),
    ("switch-chat-9", None, True),
    ("switch-workspace-1", None, True),
    ("switch-workspace-2", None, True),
    ("switch-workspace-3", None, True),
    ("switch-workspace-4", None, True),
    ("switch-workspace-5", None, True),
    ("switch-workspace-6", None, True),
    ("switch-workspace-7", None, True),
    ("switch-workspace-8", None, True),
    ("switch-workspace-9", None, True),
    ("increase-app-font-size", None, True),
    ("decrease-app-font-size", None, True),
    ("reset-app-font-size", None, True),
    ("add-workspace", "s", True),
    ("edit-workspace", "s", True),
    ("remove-workspace", "s", True),
    ("activate-workspace", "s", True),
    ("mark-workspace-as-read", "s", True),
    ("add-chat", "a{sv}", True),
    ("add-group-chat", "as", True),
]


ACCOUNT_ACTIONS = [
    ("add-contact", "a{sv}"),
    ("block-contact", "a{sv}"),
    ("remove-contact", "as"),
    ("execute-command", "(sas)"),
    ("mark-as-read", "a{sv}"),
    ("open-event", "a{sv}"),
    ("pep-config", "s"),
    ("profile", "s"),
    ("services", "s"),
    ("sync-history", "s"),
    ("subscription-accept", "a{sv}"),
    ("subscription-deny", "a{sv}"),
    ("subscription-deny-all", None),
]


ALWAYS_ACCOUNT_ACTIONS = {
    "open-event",
}


ONLINE_ACCOUNT_ACTIONS = {
    "add-contact",
    "remove-contact",
    "execute-command",
    "pep-config",
    "profile",
    "services",
    "sync-history",
    "mark-as-read",
    "subscription-accept",
    "subscription-deny",
    "subscription-deny-all",
}


FEATURE_ACCOUNT_ACTIONS = {
    "block-contact",
}
